#!/usr/bin/env python
# encoding: utf-8
#
# spectrum.py
#
# Licensed under a 3-clause BSD license.

# Revision history:
#     13 Apr 2016 J. Sánchez-Gallego
#       Initial version


from __future__ import division
from __future__ import print_function
import sys
import numpy as np
import matplotlib.pyplot as plt


class Spectrum(object):
    """A class representing an spectrum with extra functionality.

    Parameters:
        flux (array-like):
            The 1-D array contianing the spectrum.
        units (str, optional):
            The units of the flux spectrum.
        wavelength (array-like, optional):
            The wavelength solution for ``spectrum``. Must have the same number
            of elements.
        ivar (array-like, optional):
            The inverse variance array for ``spectrum``. Must have the same
            number of elements.
        mask (array-like, optional):
            The mask array for ``spectrum``. Must have the same number of
            elements.
        wavelength_unit (str, optional):
            The units of the wavelength solution.

    """

    def __init__(self, flux, units=None, wavelength_unit=None,
                 ivar=None, mask=None, wavelength=None):

        self.flux = np.array(flux)
        self.ivar = np.array(ivar) if ivar is not None else None
        self.mask = np.array(mask) if mask is not None else None
        self.wavelength = np.array(wavelength) if wavelength is not None else None

        self.units = units
        self.wavelength_unit = wavelength_unit

        # Performs some checks.

        assert len(self.flux.shape) == 1, 'spectrum must be 1-D'

        if self.ivar is not None:
            assert len(self.ivar.shape) == 1, 'ivar must be 1-D'
            assert len(self.flux) == len(self.ivar), \
                'ivar must have the same lenght as the base spectrum'

        if self.mask is not None:
            assert len(self.mask.shape) == 1, 'mask must be 1-D'
            assert len(self.flux) == len(self.mask), \
                'mask must have the same lenght as the base spectrum'

        if self.wavelength is not None:
            assert len(self.wavelength.shape) == 1, 'wavelength must be 1-D'
            assert len(self.flux) == len(self.wavelength), \
                'wavelength must have the same lenght as the base spectrum'

    def __repr__(self):
        """Representation for Spectrum."""

        return '<Marvin Spectrum ({0!s})'.format(self.flux)

    def plot(self, array='flux', xlim=None, ylim=(0, None), mask_color=None,
             xlabel=None, ylabel=None, figure=None, return_figure=False, **kwargs):
        """Plots a spectrum using matplotlib.

        Returns a |axes|_ object with a representation of this spectrum.
        The returned ``axes`` object can then be showed, modified, or saved to
        a file. If running Marvin from an iPython console and
        `matplotlib.pyplot.ion()
        <http://matplotlib.org/api/pyplot_api.html#matplotlib.pyplot.ion>`_,
        the plot will be displayed interactivelly.

        Parameters:
            array ({'flux', 'ivar', 'mask'}):
                The array to display, defaults to the internal spectrum with
                which the object was initialised.
            xlim,ylim (tuple-like or None):
                The range to display for the x- and y-axis, respectively,
                defined as a tuple of two elements ``[xmin, xmax]``. If
                the range is ``None``, the range for the axis will be set
                automatically by matploltib. If ``Spectrum.wavelength`` is
                defined, the range in the x-axis must be defined as a
                wavelength range. Default for ylim is (0, None), which cuts
                off negative values but lets the maximum float.
            xlabel,ylabel (str or None):
                The axis labels to be passed to the plot. If ``xlabel=None``
                and ``Spectrum.wavelength_unit`` is defined, those units will
                be used, after being properly formatted for Latex display.
                If ``ylabel=None``, the y-axis label will be automatically
                defined base on the type of input array.
            mask_color (matplotlib valid color or None):
                If set and ``Spectrum.mask`` is defined, the elements of
                ``array`` with ``mask`` will be coloured using that value.
                More information about `matplotlib colours
                <http://matplotlib.org/api/colors_api.html>`_.
            figure (matplotlib Figure object or None):
                The matplotlib figure object from which the axes must be
                created. If ``figure=None``, a new figure will be created.
            return_figure (bool):
                If ``True``, the matplotlib Figure object used will be returned
                along with the axes object.
            kwargs (dict):
                Any other keyword argument that will be passed to
                `matplotlib.pyplot.plot
                <http://matplotlib.org/api/pyplot_api.html#matplotlib.pyplot.plot>`_.

        Returns:
            ax:
                The `matplotlib.axes <http://matplotlib.org/api/axes_api.html>`_
                object containing the plot representing the spectrum. If
                ``return_figure=True``, a tuple will be returned of the form
                ``(ax, fig)``.

        Example:

          >>> spectrum = Spectrum(np.arange(100), wavelength=np.arange(100)*0.1)
          >>> ax = spectrum.plot(xrange=[5, 7])
          >>> ax.show()

        We can change the range of the axes after the object has been created.

          >>> ax.set_xlim(3, 8)
          >>> ax.show()

        .. |axes| replace:: matplotlib.axes
        .. _axes: http://matplotlib.org/api/axes_api.html

        """

        array = array.lower()
        validSpectrum = ['flux', 'ivar', 'mask']
        assert array in validSpectrum, 'array must be one of {0!r}'.format(validSpectrum)

        if array == 'flux':
            data = self.flux
        elif array == 'ivar':
            data = self.ivar
        elif array == 'mask':
            data = self.mask

        xaxis = self.wavelength if self.wavelength is not None else np.arange(len(self))

        fig = plt.figure() if figure is None else figure
        ax = fig.add_subplot(111)

        ax.plot(xaxis, data, **kwargs)

        # This does not work very well for small ranges of masked elements.
        # Probably requires some rethinking.
        if mask_color is not None:
            mask_indices = np.where(self.mask > 0)
            kwargs['color'] = mask_color
            ax.plot(xaxis[mask_indices], data[mask_indices], **kwargs)

        if xlim is not None:
            assert len(xlim) == 2
            ax.set_xlim(*xlim)

        if ylim is not None:
            assert len(ylim) == 2
            ax.set_ylim(*ylim)

        if xlabel is None:
            if self.wavelength is not None:
                xlabel = 'Wavelength'
                if self.wavelength_unit == 'Angstrom':
                    xlabel += r' $[\rm\AA]$'
                elif self.wavelength_unit is not None:
                    xlabel += r' [{0}]'.format(self.wavelength_unit)
            else:
                xlabel = ''

        if ylabel is None:
            if array == 'flux':
                ylabel = 'Flux'
                if self.units == '1e-17 erg/s/cm^2/Ang/spaxel':
                    ylabel += r' $[\rm 10^{-17}\,erg\,s^{-1}\,cm^{-2}\,\AA^{-1}\,spaxel^{-1}]$'
                elif self.units is not None:
                    ylabel += r' [{0}]'.format(self.units)
            elif array == 'ivar':
                ylabel = 'Inverse variance'
            else:
                ylabel = ''

        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)

        if return_figure:
            return (ax, fig)
        else:
            return ax



import json
from brain.api.base import BrainBaseView
from flask import jsonify


class BrainQueryView(BrainBaseView):
    """Class describing API calls related to queries."""

    route_base = '/query/'

    def index(self):
        self.results['data'] = 'this is a query!'
        return jsonify(self.results)




#!/usr/bin/env python
# encoding: utf-8

'''
Created by Brian Cherinka on 2016-05-23 00:13:58
Licensed under a 3-clause BSD license.

Revision History:
    Initial Version: 2016-05-23 00:13:58 by Brian Cherinka
    Last Modified On: 2016-05-23 00:13:58 by Brian

'''
from __future__ import print_function
from __future__ import division
from sqlalchemy import type_coerce
from sqlalchemy.dialects.postgresql import *


# __getitem__ broken for postgres ARRAY type
#
# this is a bug that's been fixed for 1.1.   It's detailed here:
# http://docs.sqlalchemy.org/en/latest/changelog/migration_11.html#correct-sql-types-are-established-from-indexed-access-of-array-json-hstore

# For multi-dimensional access, this can be worked around for a one-off
# using type_coerce:

#  >>> from sqlalchemy import type_coerce
#  >>> type_coerce(c[4], ARRAY(Integer))[5]

# There is also a generalized workaround created for the bug that you can
# see at
# https://bitbucket.org/zzzeek/sqlalchemy/issues/3487#comment-20200804 .
# It involves creation of an ARRAY subclass that does the right thing
# within __getitem__.   That subclass can be a drop-in replacement for
# regular ARRAY.


class ARRAY_D(ARRAY):
    class Comparator(ARRAY.Comparator):
        def __getitem__(self, index):
            super_ = super(ARRAY_D.Comparator, self).__getitem__(index)
            if not isinstance(index, slice) and self.type.dimensions > 1:
                super_ = type_coerce(
                    super_,
                    ARRAY_D(
                        self.type.item_type,
                        dimensions=self.type.dimensions - 1,
                        zero_indexes=self.type.zero_indexes)
                )
            return super_
    comparator_factory = Comparator



# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brett Andrews
# @Date:   2017-04-11


from __future__ import print_function, division, absolute_import

import pytest

from sdss_access.sync.auth import Auth


def test_set_username_interactively():
    """If authentication via `.netrc` file fails, then test for interactive prompt.

    The interactive prompt raises an IOError (merged into OSError in Python 3.3), which this test
    expects.
    """
    with pytest.raises((IOError, OSError)):
        auth = Auth(netloc='data.sdss.org', public=False)
        auth.set_username(inquire=True)



#!/usr/bin/env python
# encoding: utf-8
"""
colourPrint.py

Created by José Sánchez-Gallego on 19 Nov 2013.
Copyright (c) 2013. All rights reserved.
Licensed under a 3-clause BSD license.

This module includes astropy-based functions for colour printing.

"""

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

__ALL__ = ['colourPrint']

import sys
import multiprocessing
import threading
import codecs
import locale


try:
    get_ipython()
except NameError:
    OutStream = None
    stdio = sys
else:
    try:
        from IPython.zmq.iostream import OutStream
        from IPython.utils import io
        stdio = io
    except ImportError:
        OutStream = None
        stdio = sys


IS_PY3 = sys.version_info[0] == 3
_DEFAULT_ENCODING = 'utf-8'


def _write_with_fallback(s, write, fileobj):
    """Write the supplied string with the given write function like
    ``write(s)``, but use a writer for the locale's preferred encoding in case
    of a UnicodeEncodeError.  Failing that attempt to write with 'utf-8' or
    'latin-1'.
    """

    try:
        write(s)
        return write
    except UnicodeEncodeError:
        # Let's try the next approach...
        pass

    enc = locale.getpreferredencoding()
    try:
        Writer = codecs.getwriter(enc)
    except LookupError:
        Writer = codecs.getwriter(_DEFAULT_ENCODING)

    f = Writer(fileobj)
    write = f.write

    try:
        write(s)
        return write
    except UnicodeEncodeError:
        Writer = codecs.getwriter('latin-1')
        f = Writer(fileobj)
        write = f.write

    # If this doesn't work let the exception bubble up; I'm out of ideas
    write(s)
    return write


def isatty(file):
    """
    Returns `True` if `file` is a tty.

    Most built-in Python file-like objects have an `isatty` member,
    but some user-defined types may not, so this assumes those are not
    ttys.
    """
    if (multiprocessing.current_process().name != 'MainProcess' or
        threading.current_thread().getName() != 'MainThread'):
        return False

    if (OutStream is not None and
        isinstance(file, OutStream) and
        file.name == 'stdout'):
        return True
    elif hasattr(file, 'isatty'):
        return file.isatty()
    return False


def _color_text(text, color):
    """
    Returns a string wrapped in ANSI color codes for coloring the
    text in a terminal::

        colored_text = color_text('Here is a message', 'blue')

    This won't actually effect the text until it is printed to the
    terminal.

    Parameters
    ----------
    text : str
        The string to return, bounded by the color codes.
    color : str
        An ANSI terminal color name. Must be one of:
        black, red, green, brown, blue, magenta, cyan, lightgrey,
        default, darkgrey, lightred, lightgreen, yellow, lightblue,
        lightmagenta, lightcyan, white, or '' (the empty string).
    """
    color_mapping = {
        'black': '0;30',
        'red': '0;31',
        'green': '0;32',
        'brown': '0;33',
        'blue': '0;34',
        'magenta': '0;35',
        'cyan': '0;36',
        'lightgrey': '0;37',
        'default': '0;39',
        'darkgrey': '1;30',
        'lightred': '1;31',
        'lightgreen': '1;32',
        'yellow': '1;33',
        'lightblue': '1;34',
        'lightmagenta': '1;35',
        'lightcyan': '1;36',
        'white': '1;37'}

    if sys.platform == 'win32' and OutStream is None:
        # On Windows do not colorize text unless in IPython
        return text

    color_code = color_mapping.get(color, '0;39')
    return '\033[{0}m{1}\033[0m'.format(color_code, text)


def _decode_preferred_encoding(s):
    """Decode the supplied byte string using the preferred encoding
    for the locale (`locale.getpreferredencoding`) or, if the default encoding
    is invalid, fall back first on utf-8, then on latin-1 if the message cannot
    be decoded with utf-8.
    """

    enc = locale.getpreferredencoding()
    try:
        try:
            return s.decode(enc)
        except LookupError:
            enc = _DEFAULT_ENCODING
        return s.decode(enc)
    except UnicodeDecodeError:
        return s.decode('latin-1')


def colourPrint(*args, **kwargs):
    """
    Prints colors and styles to the terminal uses ANSI escape
    sequences.

    ::

       color_print('This is the color ', 'default', 'GREEN', 'green')

    Parameters
    ----------
    positional args : strings
        The positional arguments come in pairs (*msg*, *color*), where
        *msg* is the string to display and *color* is the color to
        display it in.

        *color* is an ANSI terminal color name.  Must be one of:
        black, red, green, brown, blue, magenta, cyan, lightgrey,
        default, darkgrey, lightred, lightgreen, yellow, lightblue,
        lightmagenta, lightcyan, white, or '' (the empty string).

    file : writeable file-like object, optional
        Where to write to.  Defaults to `sys.stdout`.  If file is not
        a tty (as determined by calling its `isatty` member, if one
        exists), no coloring will be included.

    end : str, optional
        The ending of the message.  Defaults to ``\\n``.  The end will
        be printed after resetting any color or font state.
    """

    file = kwargs.get('file', stdio.stdout)

    end = kwargs.get('end', '\n')

    write = file.write
    if isatty(file):
        for i in range(0, len(args), 2):
            msg = args[i]
            if i + 1 == len(args):
                color = ''
            else:
                color = args[i + 1]

            if color:
                msg = _color_text(msg, color)

            # Some file objects support writing unicode sensibly on some Python
            # versions; if this fails try creating a writer using the locale's
            # preferred encoding. If that fails too give up.
            if not IS_PY3 and isinstance(msg, bytes):
                msg = _decode_preferred_encoding(msg)

            write = _write_with_fallback(msg, write, file)

        write(end)
    else:
        for i in range(0, len(args), 2):
            msg = args[i]
            if not IS_PY3 and isinstance(msg, bytes):
                # Support decoding bytes to unicode on Python 2; use the
                # preferred encoding for the locale (which is *sometimes*
                # sensible)
                msg = _decode_preferred_encoding(msg)
            write(msg)
        write(end)



#!/usr/bin/env python
# encoding: utf-8
"""

test_general_pytest.py

Created by José Sánchez-Gallego on 7 Apr 2016.
Licensed under a 3-clause BSD license.

Revision history:
    7 Apr 2016 J. Sánchez-Gallego
      Initial version
    15 Jun 2016 B. Andrews
      Converted to pytest

"""

from __future__ import division
from __future__ import print_function

from collections import OrderedDict
import os

import pytest
import numpy as np
from astropy.io import fits
from astropy.wcs import WCS

import marvin
from marvin.core.core import DotableCaseInsensitive
from marvin.core.exceptions import MarvinError
from marvin.utils.general import convertCoords, get_nsa_data, getWCSFromPng, get_plot_params
from marvin.utils.dap.datamodel import get_default_plot_params


@pytest.fixture(scope='function')
def wcs(galaxy):
    return WCS(fits.getheader(galaxy.cubepath, 1))


class TestConvertCoords(object):

    def test_pix_center(self, galaxy):
        """Tests mode='pix', xyorig='center'."""

        coords = [[0, 0],
                  [5, 3],
                  [-5, 1],
                  [1, -5],
                  [10, 10],
                  [-10, -10],
                  [1.5, 2.5],
                  [0.4, 0.25]]

        expected = [[17, 17],
                    [20, 22],
                    [18, 12],
                    [12, 18],
                    [27, 27],
                    [7, 7],
                    [20, 18],
                    [17, 17]]

        cubeCoords = convertCoords(coords, mode='pix', shape=galaxy.shape)
        pytest.approx(cubeCoords, np.array(expected))

    def test_pix_lower(self, galaxy):
        """Tests mode='pix', xyorig='lower'."""

        coords = [[0, 0],
                  [5, 3],
                  [10, 10],
                  [1.5, 2.5],
                  [0.4, 0.25]]

        expected = [[0, 0],
                    [3, 5],
                    [10, 10],
                    [2, 2],
                    [0, 0]]

        cubeCoords = convertCoords(coords, mode='pix', shape=galaxy.shape,
                                   xyorig='lower')
        pytest.approx(cubeCoords, np.array(expected))

    def test_sky(self, wcs):
        """Tests mode='sky'."""

        coords = np.array([[232.5447, 48.690201],
                           [232.54259, 48.688948],
                           [232.54135, 48.692415],
                           [232.54285, 48.692372]])

        expected = [[17, 17],
                    [8, 27],
                    [33, 33],
                    [33, 26]]

        cubeCoords = convertCoords(coords, mode='sky', wcs=wcs)

        pytest.approx(cubeCoords, np.array(expected))

    @pytest.mark.parametrize('coords, mode, xyorig',
                             [([-100, 0], 'pix', 'center'),
                              ([100, 100], 'pix', 'center'),
                              ([-100, 0], 'pix', 'lower'),
                              ([100, 100], 'pix', 'lower'),
                              ([230, 48], 'sky', None),
                              ([233, 48], 'sky', None)],
                             ids=['-50_0_cen', '50_50_cen', '-50_0_low', '50_50_low',
                                  '230_48_sky', '233_48_sky'])
    def test_coords_outside_cube(self, coords, mode, xyorig, galaxy, wcs):
        kwargs = {'coords': coords, 'mode': mode}

        if xyorig is not None:
            kwargs['xyorig'] = xyorig

        if kwargs['mode'] == 'sky':
            kwargs['wcs'] = wcs

        with pytest.raises(MarvinError) as cm:
            convertCoords(shape=galaxy.shape, **kwargs)

        assert 'some indices are out of limits' in str(cm.value)


class TestGetNSAData(object):

    def _test_nsa(self, galaxy, data):
        assert isinstance(data, DotableCaseInsensitive)
        assert 'iauname' in data.keys()
        assert data['iauname'] == galaxy.iauname
        assert data['iauname'] == data.iauname
        assert 'profmean_ivar' in data.keys()
        assert 'elpetro_mag_g' in data
        assert 'version' not in data.keys()
        assert data['elpetro_mag_g'] == data.elpetro_mag_g

    def _test_drpall(self, galaxy, data):
        assert isinstance(data, DotableCaseInsensitive)
        assert 'version' in data.keys()
        assert 'profmean_ivar' not in data.keys()
        if galaxy.release != 'MPL-4':
            assert 'iauname' in data.keys()
            assert data['iauname'] == galaxy.iauname
            assert data['iauname'] == data.iauname
            assert 'elpetro_absmag' in data.keys()

    @pytest.mark.parametrize('source', [('nsa'), ('drpall')])
    def test_nsa(self, galaxy, mode, db, source):
        if mode == 'local' and source == 'nsa' and marvin.config.db is None:
            with pytest.raises(MarvinError) as cm:
                data = get_nsa_data(galaxy.mangaid, source=source, mode=mode)
            errmsg = 'get_nsa_data: cannot find a valid DB connection.'
            assert cm.type == MarvinError
            assert errmsg in str(cm.value)
        else:
            data = get_nsa_data(galaxy.mangaid, source=source, mode=mode)
            if source == 'nsa':
                self._test_nsa(galaxy, data)
            elif source == 'drpall':
                self._test_drpall(galaxy, data)

    @pytest.mark.parametrize('monkeyconfig', [('_drpall', None)], indirect=True, ids=['nodrpall'])
    def test_nodrpall(self, galaxy, monkeyconfig):
        data = get_nsa_data(galaxy.mangaid, source='drpall', mode='auto')
        self._test_drpall(galaxy, data)


class TestPillowImage(object):

    def test_image_has_wcs(self, galaxy):
        w = getWCSFromPng(galaxy.imgpath)
        assert isinstance(w, WCS) is True

    def test_use_pil(self):
        try:
            import PIL
        except ImportError as e:
            with pytest.raises(ImportError) as cm:
                err = 'No module named PIL'
                assert err == str(e.args[0])


@pytest.fixture(scope='session')
def bitmask(dapver):
    data = {'1.1.1': {'badData': {'doNotUse': 0}},
            '2.0.2': {'nocov': 0, 'badData': {'unreliable': 5, 'doNotUse': 30}}
            }
    return data[dapver]


class TestDataModelPlotParams(object):

    @pytest.mark.parametrize('name, desired',
                             [('emline_gflux', {'cmap': 'linearlab', 'percentile_clip': [5, 95], 'symmetric': False, 'snr_min': 1}),
                              ('stellar_vel', {'cmap': 'RdBu_r', 'percentile_clip': [10, 90], 'symmetric': True, 'snr_min': None}),
                              ('stellar_sigma', {'cmap': 'inferno', 'percentile_clip': [10, 90], 'symmetric': False, 'snr_min': 1})],
                             ids=['emline', 'stvel', 'stsig'])
    def test_get_plot_params(self, bitmask, dapver, name, desired):
        desired['bitmasks'] = bitmask
        actual = get_plot_params(dapver=dapver, prop=name)
        assert desired == actual






from functools import wraps
from brain.core.exceptions import BrainError

try:
    from sdss_access.path import Path
except ImportError:
    Path = None

# General Decorators
__all__ = ['parseRoutePath', 'checkPath']


def parseRoutePath(f):
    ''' Decorator to parse generic route path '''
    @wraps(f)
    def decorated_function(inst, *args, **kwargs):
        if 'path' in kwargs and kwargs['path']:
            for kw in kwargs['path'].split('/'):
                if len(kw) == 0:
                    continue
                var, value = kw.split('=')
                kwargs[var] = value
        kwargs.pop('path')
        return f(inst, *args, **kwargs)
    return decorated_function


def checkPath(func):
    '''Decorator that checks if sdss_access Path has been imported '''

    @wraps(func)
    def wrapper(*args, **kwargs):
        if not Path:
            raise BrainError('sdss_access is not installed')
        else:
            return func(*args, **kwargs)
    return wrapper



#!/usr/bin/env python
# encoding: utf-8
#
# test_modelcube.py
#
# Created by José Sánchez-Gallego on 25 Sep 2016.


from __future__ import division, print_function, absolute_import

import os

import pytest
from astropy.io import fits
from astropy.wcs import WCS

from marvin import config
from marvin.core.exceptions import MarvinError
from marvin.tools.cube import Cube
from marvin.tools.maps import Maps
from marvin.tools.modelcube import ModelCube
from marvin.tests import marvin_test_if, marvin_test_if_class


@pytest.fixture(autouse=True)
def skipmpl4(galaxy):
    if galaxy.release == 'MPL-4':
        pytest.skip('No modelcubes in MPL-4')


class TestModelCubeInit(object):

    def _test_init(self, model_cube, galaxy, bintype='SPX', template_kin='GAU-MILESHC'):
        assert model_cube._release == galaxy.release
        assert model_cube._drpver == galaxy.drpver
        assert model_cube._dapver == galaxy.dapver
        assert model_cube.bintype == bintype
        assert model_cube.template_kin == template_kin
        assert model_cube.plateifu == galaxy.plateifu
        assert model_cube.mangaid == galaxy.mangaid
        assert isinstance(model_cube.header, fits.Header)
        assert isinstance(model_cube.wcs, WCS)
        assert model_cube.wavelength is not None
        assert model_cube.redcorr is not None

    def test_init_modelcube(self, galaxy, data_origin):
        if data_origin == 'file':
            kwargs = {'filename': galaxy.modelpath}
        elif data_origin == 'db':
            kwargs = {'plateifu': galaxy.plateifu}
        elif data_origin == 'api':
            kwargs = {'plateifu': galaxy.plateifu, 'mode': 'remote'}

        model_cube = ModelCube(**kwargs)
        assert model_cube.data_origin == data_origin
        assert pytest.approx(model_cube.nsa.z, galaxy.redshift)
        self._test_init(model_cube, galaxy)

    def test_init_from_file_global_mpl4(self, galaxy):
        model_cube = ModelCube(filename=galaxy.modelpath, release='MPL-4')
        assert model_cube.data_origin == 'file'
        self._test_init(model_cube, galaxy)

    def test_raises_exception_mpl4(self, galaxy):
        with pytest.raises(MarvinError) as cm:
            ModelCube(plateifu=galaxy.plateifu, release='MPL-4')
        assert 'ModelCube requires at least dapver=\'2.0.2\'' in str(cm.value)

    def test_init_modelcube_bintype(self, galaxy, data_origin):
        kwargs = {'bintype': galaxy.bintype}
        if data_origin == 'file':
            kwargs['filename'] = galaxy.modelpath
        elif data_origin == 'db':
            kwargs['plateifu'] = galaxy.plateifu
        elif data_origin == 'api':
            kwargs['plateifu'] = galaxy.plateifu
            kwargs['mode'] = 'remote'

        model_cube = ModelCube(**kwargs)
        assert model_cube.data_origin == data_origin
        self._test_init(model_cube, galaxy, bintype=galaxy.bintype)


class TestModelCube(object):

    def test_get_flux_db(self, galaxy):
        model_cube = ModelCube(plateifu=galaxy.plateifu)
        shape = tuple([4563] + galaxy.shape)
        assert model_cube.flux.shape == shape

    def test_get_flux_api_raises_exception(self, galaxy):
        model_cube = ModelCube(plateifu=galaxy.plateifu, mode='remote')
        with pytest.raises(MarvinError) as cm:
            model_cube.flux
        assert 'cannot return a full cube in remote mode.' in str(cm.value)

    def test_get_cube_file(self, galaxy):
        model_cube = ModelCube(filename=galaxy.modelpath)
        assert isinstance(model_cube.cube, Cube)

    def test_get_maps_api(self, galaxy):
        model_cube = ModelCube(plateifu=galaxy.plateifu, mode='remote')
        assert isinstance(model_cube.maps, Maps)


@pytest.mark.slow
class TestGetSpaxel(object):

    def _test_getspaxel(self, spaxel, galaxy, bintype='SPX', template_kin='GAU-MILESHC'):
        assert spaxel._drpver == galaxy.drpver
        assert spaxel._dapver == galaxy.dapver
        assert spaxel.plateifu == galaxy.plateifu
        assert spaxel.mangaid == galaxy.mangaid
        assert spaxel.modelcube is not None
        assert spaxel.modelcube.bintype == bintype
        assert spaxel.modelcube.template_kin == template_kin
        assert spaxel._parent_shape == tuple(galaxy.shape)

        assert spaxel.model_flux is not None
        assert spaxel.model is not None
        assert spaxel.emline is not None
        assert spaxel.emline_base is not None
        assert spaxel.stellar_continuum is not None
        assert spaxel.redcorr is not None

    def test_getspaxel(self, galaxy, data_origin):
        if data_origin == 'file':
            kwargs = {'filename': galaxy.modelpath}
        elif data_origin == 'db':
            kwargs = {'plateifu': galaxy.plateifu}
        elif data_origin == 'api':
            kwargs = {'plateifu': galaxy.plateifu, 'mode': 'remote'}

        model_cube = ModelCube(**kwargs)
        spaxel = model_cube.getSpaxel(x=1, y=2)
        self._test_getspaxel(spaxel, galaxy)

    def test_getspaxel_db_api_model(self, galaxy):

        model_cube = ModelCube(plateifu=galaxy.plateifu)
        spaxel = model_cube.getSpaxel(x=1, y=2, properties=False, spectrum=False)
        self._test_getspaxel(spaxel, galaxy)
        assert spaxel.cube is None
        assert spaxel.spectrum is None
        assert spaxel.maps is None
        assert len(spaxel.properties) == 0

    def test_getspaxel_matches_file_db_remote(self, galaxy):

        modelcube_file = ModelCube(filename=galaxy.modelpath)
        modelcube_db = ModelCube(mangaid=galaxy.mangaid)
        modelcube_api = ModelCube(mangaid=galaxy.mangaid, mode='remote')

        assert modelcube_file.data_origin == 'file'
        assert modelcube_db.data_origin == 'db'
        assert modelcube_api.data_origin == 'api'

        xx = galaxy.spaxel['x']
        yy = galaxy.spaxel['y']
        idx = galaxy.spaxel['specidx']
        flux = galaxy.spaxel['model_flux']
        ivar = galaxy.spaxel['model_ivar']
        mask = galaxy.spaxel['model_mask']

        spaxel_slice_file = modelcube_file[yy, xx]
        spaxel_slice_db = modelcube_db[yy, xx]
        spaxel_slice_api = modelcube_api[yy, xx]

        assert pytest.approx(spaxel_slice_file.model_flux.flux[idx], flux)
        assert pytest.approx(spaxel_slice_db.model_flux.flux[idx], flux)
        assert pytest.approx(spaxel_slice_api.model_flux.flux[idx], flux)

        assert pytest.approx(spaxel_slice_file.model_flux.ivar[idx], ivar)
        assert pytest.approx(spaxel_slice_db.model_flux.ivar[idx], ivar)
        assert pytest.approx(spaxel_slice_api.model_flux.ivar[idx], ivar)

        assert pytest.approx(spaxel_slice_file.model_flux.mask[idx], mask)
        assert pytest.approx(spaxel_slice_db.model_flux.mask[idx], mask)
        assert pytest.approx(spaxel_slice_api.model_flux.mask[idx], mask)

        xx_cen = galaxy.spaxel['x_cen']
        yy_cen = galaxy.spaxel['y_cen']

        spaxel_getspaxel_file = modelcube_file.getSpaxel(x=xx_cen, y=yy_cen)
        spaxel_getspaxel_db = modelcube_db.getSpaxel(x=xx_cen, y=yy_cen)
        spaxel_getspaxel_api = modelcube_api.getSpaxel(x=xx_cen, y=yy_cen)

        assert pytest.approx(spaxel_getspaxel_file.model_flux.flux[idx], flux)
        assert pytest.approx(spaxel_getspaxel_db.model_flux.flux[idx], flux)
        assert pytest.approx(spaxel_getspaxel_api.model_flux.flux[idx], flux)

        assert pytest.approx(spaxel_getspaxel_file.model_flux.ivar[idx], ivar)
        assert pytest.approx(spaxel_getspaxel_db.model_flux.ivar[idx], ivar)
        assert pytest.approx(spaxel_getspaxel_api.model_flux.ivar[idx], ivar)

        assert pytest.approx(spaxel_getspaxel_file.model_flux.mask[idx], mask)
        assert pytest.approx(spaxel_getspaxel_db.model_flux.mask[idx], mask)
        assert pytest.approx(spaxel_getspaxel_api.model_flux.mask[idx], mask)


class TestPickling(object):

    def test_pickling_file(self, temp_scratch, galaxy):
        modelcube = ModelCube(filename=galaxy.modelpath, bintype=galaxy.bintype)
        assert modelcube.data_origin == 'file'
        assert isinstance(modelcube, ModelCube)
        assert modelcube.data is not None

        file = temp_scratch.join('test_modelcube.mpf')
        path = modelcube.save(str(file))

        assert file.check() is True
        assert modelcube.data is not None

        modelcube = None
        assert modelcube is None

        modelcube_restored = ModelCube.restore(str(file))
        assert modelcube_restored.data_origin == 'file'
        assert isinstance(modelcube_restored, ModelCube)
        assert modelcube_restored.data is not None

    def test_pickling_file_custom_path(self, temp_scratch, galaxy):
        modelcube = ModelCube(filename=galaxy.modelpath, bintype=galaxy.bintype)
        assert modelcube.data_origin == 'file'
        assert isinstance(modelcube, ModelCube)
        assert modelcube.data is not None

        file = temp_scratch.join('mcpickle').join('test_modelcube.mpf')
        assert file.check(file=1) is False

        path = modelcube.save(path=str(file))
        assert file.check() is True
        assert os.path.exists(path)

        modelcube_restored = ModelCube.restore(str(file), delete=True)
        assert modelcube_restored.data_origin == 'file'
        assert isinstance(modelcube_restored, ModelCube)
        assert modelcube_restored.data is not None

        assert not os.path.exists(path)

    def test_pickling_db(self, galaxy, temp_scratch):
        modelcube = ModelCube(plateifu=galaxy.plateifu, bintype=galaxy.bintype)

        file = temp_scratch.join('test_modelcube_db.mpf')
        with pytest.raises(MarvinError) as cm:
            modelcube.save(str(file))

        assert 'objects with data_origin=\'db\' cannot be saved.' in str(cm.value)

    def test_pickling_api(self, temp_scratch, galaxy):
        modelcube = ModelCube(plateifu=galaxy.plateifu, bintype=galaxy.bintype, mode='remote')
        assert modelcube.data_origin == 'api'
        assert isinstance(modelcube, ModelCube)
        assert modelcube.data is None

        file = temp_scratch.join('test_modelcube_api.mpf')
        path = modelcube.save(str(file))

        assert file.check() is True

        modelcube = None
        assert modelcube is None

        modelcube_restored = ModelCube.restore(str(file))
        assert modelcube_restored.data_origin == 'api'
        assert isinstance(modelcube_restored, ModelCube)
        assert modelcube_restored.data is None



# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2016-09-28 22:39:08
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2016-09-28 23:05:44

from __future__ import print_function, division, absolute_import


class Inspection:
    '''
    A temp alternate local class of Inspection

    Designed for Trac user authentication in the Flask web
    '''

    cols = keys = ['membername', 'modified']
    trac_url = None

    def __init__(self, session=None, username=None, auth=None):
        self.session = session
        self.message = ''
        self.status = -1
        self.set_ready()
        if self.ready:
            self.set_member(id=self.session['member_id'], update_session=False)
        if not self.ready:
            self.set_member(username=username, auth=auth)
        self.ready = True

    def set_ready(self):
        ''' sets the Flask session to ready post login '''
        if self.session is not None:
            if 'member_id' in self.session:
                try:
                    self.ready = int(self.session['member_id']) > 0
                except:
                    self.ready = False
            else:
                self.ready = False
        else:
            self.ready = False

    def set_member(self, id=None, username=None, auth=None, add=False, update_session=True):
        ''' sets the trac user member '''
        if id is None:
            id = 1 if (username, auth) == ('sdss', '43799f65a46144a0535ccea32fe2af34') else 0
        elif id == 1:
            username, auth = ('sdss', '43799f65a46144a0535ccea32fe2af34')
        if username and auth:
            self.member = {'id': int(id), 'username': username, 'auth': auth}
            fullname = "SDSS User" if id == 1 else "Unknown user"
            if self.member['id']:
                self.status = 1
                self.message = "Logged in as {0}. ".format(self.member['username'])
            else:
                self.status = 0
                self.message = "Failed login for {0}. ".format(username)
                self.message += "Please retry." if username == 'sdss' else "Username unrecognized."
        else:
            self.member = None
        if self.member and update_session:
            self.set_session_member(id=self.member['id'],
                                    username=self.member['username'],
                                    fullname=fullname, auth=self.member['auth'])

    def set_session_member(self, id=None, username=None, fullname=None, auth=None):
        ''' sets the trac user member into the Flask session '''
        if self.session is not None:
            if 'member_id' not in self.session or self.session['member_id'] != id:
                try:
                    self.session['member_id'] = int(id)
                    self.session['member_username'] = username if username else 'None'
                    self.session['member_fullname'] = fullname if fullname else 'None'
                    self.session['member_auth'] = auth if auth else 'None'
                    self.set_ready()
                except:
                    self.session['member_id'] = None
                    self.session['member_username'] = None
                    self.session['member_fullname'] = None
                    self.session['member_auth'] = None

    def set_member_from_session(self):
        ''' sets the trac user member from a Flask session '''
        if self.session is not None:
            member_id = self.session['member_id'] if 'member_id' in self.session else None
            member_username = self.session['member_username'] if 'member_username' in self.session else None
            member_auth = self.session['member_auth'] if 'member_auth' in self.session else None
            self.set_member(id=member_id, username=member_username, auth=member_auth, update_session=False)
        else:
            self.member = None

    def result(self):
        result = {'ready': self.ready, 'message': self.message, 'status': self.status}
        if self.session and 'member_fullname' in self.session:
            result.update({'membername': self.session['member_fullname']})
        return result



#!/usr/bin/env python
# encoding: utf-8

'''
Created by Brian Cherinka on 2016-03-28 23:30:14
Licensed under a 3-clause BSD license.

Revision History:
    Initial Version: 2016-03-28 23:30:14 by Brian Cherinka
    Last Modified On: 2016-03-28 23:30:14 by Brian

'''
from __future__ import print_function
from __future__ import division
from brain.db.modelGraph import ModelGraph
from marvin import config, log
import inspect

__author__ = 'Brian Cherinka'


class MarvinDB(object):
    ''' Class designed to handle database related things with Marvin '''

    def __init__(self, *args, **kwargs):
        self.dbtype = kwargs.get('dbtype', None)
        self.db = None
        self.log = kwargs.get('log', None)
        self.error = []
        self.__init_the_db()

    def __init_the_db(self):
        ''' Initialize the db '''
        if self.dbtype:
            self._setupDB()
        if self.db:
            self._importModels()
        else:
            self.datadb = None
            self.sampledb = None
            self.dapdb = None
        self._setSession()
        self.testDbConnection()
        self._setModelGraph()
        self.cache_bits = []
        if self.db:
            self._addCache()

    def _setupDB(self):
        ''' Try to import the database '''
        try:
            from marvin.db.database import db
        except RuntimeError as e:
            log.debug('RuntimeError raised: Problem importing db: {0}'.format(e))
            self.db = None
        except ImportError as e:
            log.debug('ImportError raised: Problem importing db: {0}'.format(e))
            self.db = None
        else:
            self.db = db

    def _importModels(self):
        ''' Try to import the sql alchemy model classes '''

        try:
            import marvin.db.models.SampleModelClasses as sampledb
        except Exception as e:
            log.debug('Exception raised: Problem importing mangadb SampleModelClasses: {0}'.format(e))
            self.sampledb = None
        else:
            self.sampledb = sampledb

        try:
            import marvin.db.models.DataModelClasses as datadb
        except Exception as e:
            log.debug('Exception raised: Problem importing mangadb DataModelClasses: {0}'.format(e))
            self.datadb = None
        else:
            self.datadb = datadb

        try:
            import marvin.db.models.DapModelClasses as dapdb
        except Exception as e:
            log.debug('Exception raised: Problem importing mangadb DapModelClasses: {0}'.format(e))
            self.dapdb = None
            self.spaxelpropdict = None
        else:
            self.dapdb = dapdb
            self.spaxelpropdict = self._setSpaxelPropDict()

    def _setSpaxelPropDict(self):
        ''' Set the SpaxelProp lookup dictionary '''
        newmpls = [m for m in config._mpldict.keys() if m > 'MPL-4']
        spdict = {'MPL-4': 'SpaxelProp'}
        newdict = {mpl: 'SpaxelProp{0}'.format(mpl.split('-')[1]) for mpl in newmpls}
        spdict.update(newdict)
        return spdict

    def _getSpaxelProp(self):
        ''' Get the correct SpaxelProp class given an MPL '''
        return {'full': self.spaxelpropdict[self._release], 'clean':
                'Clean{0}'.format(self.spaxelpropdict[self._release])}

    def _setSession(self):
        ''' Sets the database session '''
        self.session = self.db.Session() if self.db else None

    def testDbConnection(self):
        ''' Test the database connection to ensure it works.  Sets a boolean variable isdbconnected '''
        if self.db and self.datadb:
            try:
                tmp = self.session.query(self.datadb.PipelineVersion).first()
            except Exception as e:
                self.isdbconnected = False
                self.error.append('Error connecting to manga database: {0}'.format(str(e)))
            else:
                self.isdbconnected = True
        else:
            self.isdbconnected = False

    def forceDbOff(self):
        ''' Force the database to turn off '''
        self.db = None
        self.session = None
        self.isdbconnected = False
        self.datadb = None
        self.dapdb = None
        self.sampledb = None

    def forceDbOn(self, dbtype=None):
        ''' Force the database to turn on '''
        self.__init_the_db()

    def generateClassDict(self, module=None, lower=None):
        ''' Generates a dictionary of the Model Classes, based on class name as key, to the object class.
            Selects only those classes in the module with attribute __tablename__
            lower = True makes class name key all lowercase
        '''
        if not module:
            module = self.datadb

        classdict = {}
        for model in inspect.getmembers(module, inspect.isclass):
            keyname = model[0].lower() if lower else model[0]
            if hasattr(model[1], '__tablename__'):
                # only include the spaxelprop table matching the MPL version
                if 'SpaxelProp' in keyname:
                    if keyname in self._getSpaxelProp().values():
                        classdict[keyname] = model[1]
                else:
                    classdict[keyname] = model[1]
        return classdict

    def buildUberClassDict(self, **kwargs):
        ''' Builds an uber class dictionary from all modelclasses '''
        self._release = kwargs.get('release', config.release)
        classdict = {}
        models = [self.datadb, self.sampledb, self.dapdb]
        for model in models:
            if model:
                modelclasses = self.generateClassDict(module=model)
                classdict.update(modelclasses)
        return classdict

    def _setModelGraph(self):
        ''' Initiates the ModelGraph using all available ModelClasses '''
        models = list(filter(None, [self.datadb, self.sampledb, self.dapdb]))
        if models:
            self.modelgraph = ModelGraph(models)
        else:
            self.modelgraph = None

    def _addCache(self):
        ''' Initialize dogpile caching for relationships

        Caching options.   A set of three RelationshipCache options
        which can be applied to Query(), causing the "lazy load"
        of these attributes to be loaded from cache.

        '''

        if self.datadb:
            self.cache_bits.append(self.datadb.data_cache)

        if self.sampledb:
            self.cache_bits.append(self.sampledb.sample_cache)

        if self.dapdb:
            self.cache_bits.append(self.dapdb.dap_cache)



import sqlalchemy as sa
from wtforms import Form
from wtforms.compat import text_type

from wtforms_alchemy import (
    GroupedQuerySelectField,
    QuerySelectField,
    QuerySelectMultipleField
)


class DummyPostData(dict):
    def getlist(self, key):
        v = self[key]
        if not isinstance(v, (list, tuple)):
            v = [v]
        return v


class LazySelect(object):
    def __call__(self, field, **kwargs):
        return list(
            (val, text_type(label), selected)
            for val, label, selected in field.iter_choices()
        )


class Base(object):
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)


class TestBase(object):
    def _do_tables(self, mapper, engine):
        metadata = sa.MetaData()

        test_table = sa.Table(
            'test', metadata,
            sa.Column('id', sa.Integer, primary_key=True, nullable=False),
            sa.Column('name', sa.String, nullable=False),
        )

        pk_test_table = sa.Table(
            'pk_test', metadata,
            sa.Column('foobar', sa.String, primary_key=True, nullable=False),
            sa.Column('baz', sa.String, nullable=False),
        )

        Test = type(str('Test'), (Base, ), {})
        PKTest = type(str('PKTest'), (Base, ), {
            '__unicode__': lambda x: x.baz,
            '__str__': lambda x: x.baz,
        })

        mapper(Test, test_table, order_by=[test_table.c.name])
        mapper(PKTest, pk_test_table, order_by=[pk_test_table.c.baz])
        self.Test = Test
        self.PKTest = PKTest

        metadata.create_all(bind=engine)

    def _fill(self, sess):
        for i, n in [(1, 'apple'), (2, 'banana')]:
            s = self.Test(id=i, name=n)
            p = self.PKTest(foobar='hello%s' % (i, ), baz=n)
            sess.add(s)
            sess.add(p)
        sess.flush()
        sess.commit()


class TestQuerySelectField(TestBase):
    def setup_method(self, method):
        engine = sa.create_engine('sqlite:///:memory:', echo=False)
        self.Session = sa.orm.session.sessionmaker(bind=engine)
        from sqlalchemy.orm import mapper
        self._do_tables(mapper, engine)

    def test_without_factory(self):
        sess = self.Session()
        self._fill(sess)

        class F(Form):
            a = QuerySelectField(
                get_label='name',
                widget=LazySelect(),
                get_pk=lambda x: x.id
            )
        form = F(DummyPostData(a=['1']))
        form.a.query = sess.query(self.Test)
        assert form.a.data is not None
        assert form.a.data.id, 1
        assert form.a(), [('1', 'apple', True), ('2', 'banana', False)]
        assert form.validate()

        form = F(a=sess.query(self.Test).filter_by(name='banana').first())
        form.a.query = sess.query(self.Test).filter(self.Test.name != 'banana')
        assert not form.validate()
        assert form.a.errors, ['Not a valid choice']

    def test_with_query_factory(self):
        sess = self.Session()
        self._fill(sess)

        class F(Form):
            a = QuerySelectField(
                get_label=(lambda model: model.name),
                query_factory=lambda: sess.query(self.Test),
                widget=LazySelect()
            )
            b = QuerySelectField(
                allow_blank=True,
                query_factory=lambda: sess.query(self.PKTest),
                widget=LazySelect()
            )

        form = F()
        assert form.a.data is None
        assert form.a() == [('1', 'apple', False), ('2', 'banana', False)]
        assert form.b.data is None
        assert form.b() == [
            ('__None', '', True),
            ('hello1', 'apple', False),
            ('hello2', 'banana', False)
        ]
        assert not form.validate()

        form = F(DummyPostData(a=['1'], b=['hello2']))
        assert form.a.data.id == 1
        assert form.a() == [('1', 'apple', True), ('2', 'banana', False)]
        assert form.b.data.baz == 'banana'
        assert form.b() == [
            ('__None', '', False),
            ('hello1', 'apple', False),
            ('hello2', 'banana', True)
        ]
        assert form.validate()

        # Make sure the query is cached
        sess.add(self.Test(id=3, name='meh'))
        sess.flush()
        sess.commit()
        assert form.a() == [('1', 'apple', True), ('2', 'banana', False)]
        form.a._object_list = None
        assert form.a() == [
            ('1', 'apple', True), ('2', 'banana', False), ('3', 'meh', False)
        ]

        # Test bad data
        form = F(DummyPostData(b=['__None'], a=['fail']))
        assert not form.validate()
        assert form.a.errors == ['Not a valid choice']
        assert form.b.errors == []
        assert form.b.data is None


class TestQuerySelectMultipleField(TestBase):
    def setup_method(self, method):
        from sqlalchemy.orm import mapper
        engine = sa.create_engine('sqlite:///:memory:', echo=False)
        Session = sa.orm.session.sessionmaker(bind=engine)
        self._do_tables(mapper, engine)
        self.sess = Session()
        self._fill(self.sess)

    class F(Form):
        a = QuerySelectMultipleField(get_label='name', widget=LazySelect())

    def test_unpopulated_default(self):
        form = self.F()
        assert [] == form.a.data

    def test_single_value_without_factory(self):
        form = self.F(DummyPostData(a=['1']))
        form.a.query = self.sess.query(self.Test)
        assert [1] == [v.id for v in form.a.data]
        assert form.a() == [('1', 'apple', True), ('2', 'banana', False)]
        assert form.validate()

    def test_multiple_values_without_query_factory(self):
        form = self.F(DummyPostData(a=['1', '2']))
        form.a.query = self.sess.query(self.Test)
        assert [1, 2] == [v.id for v in form.a.data]
        assert form.a() == [('1', 'apple', True), ('2', 'banana', True)]
        assert form.validate()

        form = self.F(DummyPostData(a=['1', '3']))
        form.a.query = self.sess.query(self.Test)
        assert [x.id for x in form.a.data], [1]
        assert not form.validate()

    def test_single_default_value(self):
        first_test = self.sess.query(self.Test).get(2)

        class F(Form):
            a = QuerySelectMultipleField(
                get_label='name',
                default=[first_test],
                widget=LazySelect(),
                query_factory=lambda: self.sess.query(self.Test)
            )
        form = F()
        assert [v.id for v in form.a.data], [2]
        assert form.a(), [('1', 'apple', False), ('2', 'banana', True)]
        assert form.validate()


class DatabaseTestCase(object):
    def setup_method(self, method):
        self.engine = sa.create_engine('sqlite:///:memory:')

        self.base = sa.ext.declarative.declarative_base()
        self.create_models()

        self.base.metadata.create_all(self.engine)

        Session = sa.orm.session.sessionmaker(bind=self.engine)
        self.session = Session()

    def teardown_method(self, method):
        self.session.close_all()
        self.base.metadata.drop_all(self.engine)
        self.engine.dispose()


class TestGroupedQuerySelectField(DatabaseTestCase):
    def create_models(self):
        class City(self.base):
            __tablename__ = 'city'
            id = sa.Column(sa.Integer, primary_key=True)
            name = sa.Column(sa.String)
            country = sa.Column(sa.String)

        self.City = City

    def create_cities(self):
        self.session.add_all([
            self.City(name='Helsinki', country='Finland'),
            self.City(name='Vantaa', country='Finland'),
            self.City(name='New York', country='USA'),
            self.City(name='Washington', country='USA'),
            self.City(name='Stockholm', country='Sweden'),
        ])

    def create_form(self, **kwargs):
        query = self.session.query(self.City).order_by('name', 'country')

        class MyForm(Form):
            city = GroupedQuerySelectField(
                label=kwargs.get('label', 'City'),
                query_factory=kwargs.get('query_factory', lambda: query),
                get_label=kwargs.get('get_label', lambda c: c.name),
                get_group=kwargs.get('get_group', lambda c: c.country),
                allow_blank=kwargs.get('allow_blank', False),
                blank_text=kwargs.get('blank_text', ''),
                blank_value=kwargs.get('blank_value', '__None'),
            )

        return MyForm

    def test_rendering(self):
        MyForm = self.create_form()
        self.create_cities()
        assert str(MyForm().city).replace('\n', '') == (
            '<select id="city" name="city">'
            '<optgroup label="Finland">'
            '<option value="1">Helsinki</option>'
            '<option value="2">Vantaa</option>'
            '</optgroup><optgroup label="Sweden">'
            '<option value="5">Stockholm</option>'
            '</optgroup>'
            '<optgroup label="USA">'
            '<option value="3">New York</option>'
            '<option value="4">Washington</option>'
            '</optgroup>'
            '</select>'
        )

    def test_custom_none_value(self):
        self.create_cities()
        MyForm = self.create_form(
            allow_blank=True,
            blank_text='Choose city...',
            blank_value=''
        )
        form = MyForm(DummyPostData({'city': ''}))
        assert form.validate(), form.errors
        assert '<option selected value="">Choose city...</option>' in (
            str(form.city)
        )



# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-06-12 19:00:10
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2017-06-12 19:13:15

from __future__ import print_function, division, absolute_import
from marvin.core.core import Dotable, DotableCaseInsensitive
import pytest

from collections import OrderedDict

normal_dicts = [OrderedDict(A=7, b=[10, 2], C='AbCdEf', d=['ghIJ', 'Lmnop'])]


@pytest.fixture(params=normal_dicts)
def dotdict(request):
    return Dotable(request.param)


@pytest.fixture(params=normal_dicts)
def dotdictci(request):
    return DotableCaseInsensitive(request.param)


class TestCoreDotable(object):

    @pytest.mark.parametrize('key', [('A'), ('b'), ('C'), ('d')])
    def test_dotable(self, dotdict, key):
        assert dotdict[key] == dotdict.__getattr__(key)

    @pytest.mark.parametrize('key', [('A'), ('b'), ('C'), ('d')])
    def test_dotablecaseins(self, dotdictci, key):
        assert dotdictci[key] == dotdictci.__getattr__(key)
        assert dotdictci[key.upper()] == dotdictci.__getattr__(key.lower())
        assert dotdictci[key.lower()] == dotdictci.__getattr__(key.upper())
        assert dotdictci[key.lower()] == dotdictci.__getattr__(key.lower())



#!/usr/bin/env python
# encoding: utf-8
#
# maps.py
#
# Created by José Sánchez-Gallego on 20 Jun 2016.


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import distutils.version
import warnings
import itertools

import astropy.io.fits
import astropy.wcs
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

import marvin
import marvin.api.api
import marvin.core.core
import marvin.core.exceptions
import marvin.tools.cube
import marvin.tools.map
import marvin.tools.spaxel
import marvin.utils.general.general
import marvin.utils.dap
import marvin.utils.dap.bpt
import marvin.utils.six

try:
    import sqlalchemy
except ImportError:
    sqlalchemy = None


# The values in the bintypes dictionary for MPL-4 are the execution plan id
# for each bintype.
__BINTYPES_MPL4__ = {'NONE': 3, 'RADIAL': 7, 'STON': 1}
__BINTYPES_MPL4_UNBINNED__ = 'NONE'
__BINTYPES__ = ['ALL', 'NRE', 'SPX', 'VOR10']
__BINTYPES_UNBINNED__ = 'SPX'

__TEMPLATES_KIN_MPL4__ = ['M11-STELIB-ZSOL', 'MIUSCAT-THIN', 'MILES-THIN']
__TEMPLATES_KIN_MPL4_DEFAULT__ = 'MIUSCAT-THIN'
__TEMPLATES_KIN__ = ['GAU-MILESHC']
__TEMPLATES_KIN_DEFAULT__ = 'GAU-MILESHC'

__all__ = ('Maps')


def _is_MPL4(dapver):
    """Returns True if the dapver version is <= MPL-4."""

    assert isinstance(dapver, marvin.utils.six.string_types), 'dapver must be a string'

    if 'v' in dapver:
        dapver = dapver.strip('v').replace('_', '.')

    dap_version = distutils.version.StrictVersion(dapver)
    MPL4_version = distutils.version.StrictVersion('1.1.1')

    return dap_version <= MPL4_version


def _get_bintype(dapver, bintype=None):
    """Checks the bintype and returns the default value if None."""

    if bintype is not None:
        bintype = bintype.upper()
        bintypes_check = __BINTYPES_MPL4__.keys() if _is_MPL4(dapver) else __BINTYPES__
        assert bintype in bintypes_check, ('invalid bintype. bintype must be one of {0}'
                                           .format(bintypes_check))
        return bintype

    # Defines the default value depending on the version
    if _is_MPL4(dapver):
        return __BINTYPES_MPL4_UNBINNED__
    else:
        return __BINTYPES_UNBINNED__


def _get_template_kin(dapver, template_kin=None):
    """Checks the template_kin and returns the default value if None."""

    if template_kin is not None:
        template_kin = template_kin.upper()
        templates_check = __TEMPLATES_KIN_MPL4__ if _is_MPL4(dapver) else __TEMPLATES_KIN__
        assert template_kin in templates_check, ('invalid template_kin. '
                                                 'template_kin must be one of {0}'
                                                 .format(templates_check))
        return template_kin

    # Defines the default value depending on the version
    if _is_MPL4(dapver):
        return __TEMPLATES_KIN_MPL4_DEFAULT__
    else:
        return __TEMPLATES_KIN_DEFAULT__


def _get_bintemps(dapver, default=None):
    ''' Get a list of all bin-template types for a given MPL '''

    if _is_MPL4(dapver):
        bins = __BINTYPES_MPL4__.keys()
        temps = __TEMPLATES_KIN_MPL4__
    else:
        bins = __BINTYPES__
        temps = __TEMPLATES_KIN__

    bintemps = ['-'.join(item) for item in list(itertools.product(bins, temps))]

    if default:
        bintemps = '{0}-{1}'.format(_get_bintype(dapver), _get_template_kin(dapver))

    return bintemps


class Maps(marvin.core.core.MarvinToolsClass):
    """Returns an object representing a DAP Maps file.

    Parameters:
        data (``HDUList``, SQLAlchemy object, or None):
            An astropy ``HDUList`` or a SQLAlchemy object of a maps, to
            be used for initialisation. If ``None``, the normal mode will
            be used (see :ref:`mode-decision-tree`).
        filename (str):
            The path of the data cube file containing the spaxel to load.
        mangaid (str):
            The mangaid of the spaxel to load.
        plateifu (str):
            The plate-ifu of the spaxel to load (either ``mangaid`` or
            ``plateifu`` can be used, but not both).
        mode ({'local', 'remote', 'auto'}):
            The load mode to use. See :ref:`mode-decision-tree`.
        bintype (str or None):
            The binning type. For MPL-4, one of the following: ``'NONE',
            'RADIAL', 'STON'`` (if ``None`` defaults to ``'NONE'``).
            For MPL-5 and successive, one of, ``'ALL', 'NRE', 'SPX', 'VOR10'``
            (defaults to ``'SPX'``).
        template_kin (str or None):
            The template use for kinematics. For MPL-4, one of
            ``'M11-STELIB-ZSOL', 'MILES-THIN', 'MIUSCAT-THIN'`` (if ``None``,
            defaults to ``'MIUSCAT-THIN'``). For MPL-5 and successive, the only
            option in ``'GAU-MILESHC'`` (``None`` defaults to it).
        template_pop (str or None):
            A placeholder for a future version in which stellar populations
            are fitted using a different template that ``template_kin``. It
            has no effect for now.
        nsa_source ({'auto', 'drpall', 'nsa'}):
            Defines how the NSA data for this object should loaded when
            ``Maps.nsa`` is first called. If ``drpall``, the drpall file will
            be used (note that this will only contain a subset of all the NSA
            information); if ``nsa``, the full set of data from the DB will be
            retrieved. If the drpall file or a database are not available, a
            remote API call will be attempted. If ``nsa_source='auto'``, the
            source will depend on how the ``Maps`` object has been
            instantiated. If the cube has ``Maps.data_origin='file'``,
            the drpall file will be used (as it is more likely that the user
            has that file in their system). Otherwise, ``nsa_source='nsa'``
            will be assumed. This behaviour can be modified during runtime by
            modifying the ``Maps.nsa_mode`` with one of the valid values.
        release (str):
            The MPL/DR version of the data to use.

    """

    def __init__(self, *args, **kwargs):

        valid_kwargs = [
            'data', 'filename', 'mangaid', 'plateifu', 'mode', 'release',
            'bintype', 'template_kin', 'template_pop', 'nsa_source']

        assert len(args) == 0, 'Maps does not accept arguments, only keywords.'
        for kw in kwargs:
            assert kw in valid_kwargs, 'keyword {0} is not valid'.format(kw)

        # For now, we set bintype and template_kin to the kwarg values, so that
        # they can be used by getFullPath.
        self.bintype = kwargs.get('bintype', None)
        self.template_kin = kwargs.get('template_kin', None)

        super(Maps, self).__init__(*args, **kwargs)

        if kwargs.pop('template_pop', None):
            warnings.warn('template_pop is not yet in use. Ignoring value.',
                          marvin.core.exceptions.MarvinUserWarning)

        # We set the bintype  and template_kin again, now using the DAP version
        self.bintype = _get_bintype(self._dapver, bintype=kwargs.pop('bintype', None))
        self.template_kin = _get_template_kin(self._dapver,
                                              template_kin=kwargs.pop('template_kin', None))
        self.template_pop = None

        self.header = None
        self.wcs = None
        self.shape = None
        self._cube = None

        if self.data_origin == 'file':
            self._load_maps_from_file(data=self.data)
        elif self.data_origin == 'db':
            self._load_maps_from_db(data=self.data)
        elif self.data_origin == 'api':
            self._load_maps_from_api()
        else:
            raise marvin.core.exceptions.MarvinError(
                'data_origin={0} is not valid'.format(self.data_origin))

        self.properties = marvin.utils.dap.get_dap_datamodel(self._dapver)

        self._check_versions(self)

    def __repr__(self):
        return ('<Marvin Maps (plateifu={0.plateifu!r}, mode={0.mode!r}, '
                'data_origin={0.data_origin!r}, bintype={0.bintype}, '
                'template_kin={0.template_kin})>'.format(self))

    def __getitem__(self, value):
        """Gets either a spaxel or a map depending on the type on input."""

        if isinstance(value, tuple):
            assert len(value) == 2, 'slice must have two elements.'
            y, x = value
            return self.getSpaxel(x=x, y=y, xyorig='lower')
        elif isinstance(value, marvin.utils.six.string_types):
            parsed_property = self.properties.get(value)
            if parsed_property is None:
                raise marvin.core.exceptions.MarvinError('invalid property')
            maps_property, channel = parsed_property
            return self.getMap(maps_property.name, channel=channel)
        else:
            raise marvin.core.exceptions.MarvinError('invalid type for getitem.')

    @staticmethod
    def _check_versions(instance):
        """Confirm that drpver and dapver match the ones from the header.

        This is written as a staticmethod because we'll also use if for
        ModelCube.

        """

        header_drpver = instance.header['VERSDRP3']
        isMPL4 = False
        if instance._release == 'MPL-4' and header_drpver == 'v1_5_0':
            header_drpver = 'v1_5_1'
            isMPL4 = True
        assert header_drpver == instance._drpver, ('mismatch between maps._drpver={0} '
                                                   'and header drpver={1}'
                                                   .format(instance._drpver, header_drpver))

        # MPL-4 does not have VERSDAP
        if isMPL4:
            assert 'VERSDAP' not in instance.header, 'mismatch between maps._dapver and header'
        else:
            header_dapver = instance.header['VERSDAP']
            assert header_dapver == instance._dapver, 'mismatch between maps._dapver and header'

    def _getFullPath(self):
        """Returns the full path of the file in the tree."""

        params = self._getPathParams()
        path_type = params.pop('path_type')

        return super(Maps, self)._getFullPath(path_type, **params)

    def download(self):
        """Downloads the maps using sdss_access - Rsync"""

        if not self.plateifu:
            return None

        params = self._getPathParams()
        path_type = params.pop('path_type')

        return super(Maps, self).download(path_type, **params)

    def _getPathParams(self):
        """Returns a dictionary with the paramters of the Maps file.

        The output of this class is mostly intended to be used by
        :func:`Maps._getFullPath` and :func:`Maps.download`.

        """

        plate, ifu = self.plateifu.split('-')

        bintype = _get_bintype(self._dapver, bintype=self.bintype)
        template_kin = _get_template_kin(self._dapver, template_kin=self.template_kin)

        if _is_MPL4(self._dapver):
            niter = int('{0}{1}'.format(__TEMPLATES_KIN_MPL4__.index(template_kin),
                                        __BINTYPES_MPL4__[bintype]))
            params = dict(drpver=self._drpver, dapver=self._dapver,
                          plate=plate, ifu=ifu, bintype=bintype, n=niter,
                          path_type='mangamap')
        else:
            daptype = '{0}-{1}'.format(bintype, template_kin)
            params = dict(drpver=self._drpver, dapver=self._dapver,
                          plate=plate, ifu=ifu, mode='MAPS', daptype=daptype,
                          path_type='mangadap5')

        return params

    def _load_maps_from_file(self, data=None):
        """Loads a MAPS file."""

        if data is not None:
            assert isinstance(data, astropy.io.fits.HDUList), 'data is not a HDUList.'
        else:
            self.data = astropy.io.fits.open(self.filename)

        self.header = self.data[0].header

        self.mangaid = self.header['MANGAID'].strip()
        self.plateifu = self.header['PLATEIFU'].strip()

        # We use EMLINE_GFLUX because is present in MPL-4 and 5 and is not expected to go away.
        header = self.data['EMLINE_GFLUX'].header
        naxis = header['NAXIS']
        wcs_pre = astropy.wcs.WCS(header)
        # Takes only the first two axis.
        self.wcs = wcs_pre.sub(2) if naxis > 2 else naxis
        self.shape = self.data['EMLINE_GFLUX'].data.shape[-2:]

        # Checks and populates release.
        file_drpver = self.header['VERSDRP3']
        file_drpver = 'v1_5_1' if file_drpver == 'v1_5_0' else file_drpver

        file_ver = marvin.config.lookUpRelease(file_drpver)
        assert file_ver is not None, 'cannot find file version.'

        if file_ver != self._release:
            warnings.warn('mismatch between file version={0} and object release={1}. '
                          'Setting object release to {0}'.format(file_ver, self._release),
                          marvin.core.exceptions.MarvinUserWarning)
            self._release = file_ver

        self._drpver, self._dapver = marvin.config.lookUpVersions(release=self._release)

        # Checks the bintype and template_kin from the header
        if not _is_MPL4(self._dapver):
            header_bintype = self.data[0].header['BINKEY'].strip().upper()
            header_bintype = 'SPX' if header_bintype == 'NONE' else header_bintype
        else:
            header_bintype = self.data[0].header['BINTYPE'].strip().upper()

        header_template_kin_key = 'TPLKEY' if _is_MPL4(self._dapver) else 'SCKEY'
        header_template_kin = self.data[0].header[header_template_kin_key].strip().upper()

        if self.bintype != header_bintype:
            self.bintype = header_bintype

        if self.template_kin != header_template_kin:
            self.template_kin = header_template_kin

    def _load_maps_from_db(self, data=None):
        """Loads the ``mangadap.File`` object for this Maps."""

        mdb = marvin.marvindb

        plate, ifu = self.plateifu.split('-')

        if not mdb.isdbconnected:
            raise marvin.core.exceptions.MarvinError('No db connected')

        if sqlalchemy is None:
            raise marvin.core.exceptions.MarvinError('sqlalchemy required to access the local DB.')

        if data is not None:
            assert isinstance(data, mdb.dapdb.File), 'data in not a marvindb.dapdb.File object.'
        else:

            datadb = mdb.datadb
            dapdb = mdb.dapdb
            # Initial query for version
            version_query = mdb.session.query(dapdb.File).join(
                datadb.PipelineInfo,
                datadb.PipelineVersion).filter(
                    datadb.PipelineVersion.version == self._dapver).from_self()

            # Query for maps parameters
            db_maps_file = version_query.join(
                datadb.Cube,
                datadb.IFUDesign).filter(
                    datadb.Cube.plate == plate,
                    datadb.IFUDesign.name == str(ifu)).from_self().join(
                        dapdb.FileType).filter(dapdb.FileType.value == 'MAPS').join(
                            dapdb.Structure, dapdb.BinType).join(
                                dapdb.Template,
                                dapdb.Structure.template_kin_pk == dapdb.Template.pk).filter(
                                    dapdb.BinType.name == self.bintype,
                                    dapdb.Template.name == self.template_kin).all()

            if len(db_maps_file) > 1:
                raise marvin.core.exceptions.MarvinError(
                    'more than one Maps file found for this combination of parameters.')
            elif len(db_maps_file) == 0:
                raise marvin.core.exceptions.MarvinError(
                    'no Maps file found for this combination of parameters.')

            self.data = db_maps_file[0]

        self.header = self.data.primary_header

        # Gets the cube header
        cubehdr = self.data.cube.header

        # Gets the mangaid
        self.mangaid = cubehdr['MANGAID'].strip()

        # Gets the shape from the associated cube.
        self.shape = self.data.cube.shape.shape

        # Creates the WCS from the cube's WCS header
        self.wcs = astropy.wcs.WCS(self.data.cube.wcs.makeHeader())

    def _load_maps_from_api(self):
        """Loads a Maps object from remote."""

        url = marvin.config.urlmap['api']['getMaps']['url']

        url_full = url.format(name=self.plateifu,
                              bintype=self.bintype,
                              template_kin=self.template_kin)

        try:
            response = self._toolInteraction(url_full)
        except Exception as ee:
            raise marvin.core.exceptions.MarvinError(
                'found a problem when checking if remote maps exists: {0}'.format(str(ee)))

        data = response.getData()

        if self.plateifu not in data['plateifu']:
            raise marvin.core.exceptions.MarvinError('remote maps has a different plateifu!')

        self.header = astropy.io.fits.Header.fromstring(data['header'])

        # Sets the mangaid
        self.mangaid = data['mangaid']

        # Gets the shape from the associated cube.
        self.shape = data['shape']

        # Sets the WCS
        self.wcs = astropy.wcs.WCS(astropy.io.fits.Header.fromstring(data['wcs']))

        return

    @property
    def cube(self):
        """Returns the :class:`~marvin.tools.cube.Cube` for with this Maps."""

        if not self._cube:
            try:
                cube = marvin.tools.cube.Cube(plateifu=self.plateifu,
                                              release=self._release)
            except Exception as err:
                raise marvin.core.exceptions.MarvinError(
                    'cannot instantiate a cube for this Maps. Error: {0}'.format(err))
            self._cube = cube

        return self._cube

    def getSpaxel(self, x=None, y=None, ra=None, dec=None,
                  spectrum=True, modelcube=False, **kwargs):
        """Returns the |spaxel| matching certain coordinates.

        The coordinates of the spaxel to return can be input as ``x, y`` pixels
        relative to``xyorig`` in the cube, or as ``ra, dec`` celestial
        coordinates.

        If ``spectrum=True``, the returned |spaxel| will be instantiated with the
        DRP spectrum of the spaxel for the DRP cube associated with this Maps.

        Parameters:
            x,y (int or array):
                The spaxel coordinates relative to ``xyorig``. If ``x`` is an
                array of coordinates, the size of ``x`` must much that of
                ``y``.
            ra,dec (float or array):
                The coordinates of the spaxel to return. The closest spaxel to
                those coordinates will be returned. If ``ra`` is an array of
                coordinates, the size of ``ra`` must much that of ``dec``.
            xyorig ({'center', 'lower'}):
                The reference point from which ``x`` and ``y`` are measured.
                Valid values are ``'center'`` (default), for the centre of the
                spatial dimensions of the cube, or ``'lower'`` for the
                lower-left corner. This keyword is ignored if ``ra`` and
                ``dec`` are defined.
            spectrum (bool):
                If ``True``, the |spaxel| will be initialised with the
                corresponding DRP spectrum.
            modelcube (bool):
                If ``True``, the |spaxel| will be initialised with the
                corresponding ModelCube data.


        Returns:
            spaxels (list):
                The |spaxel| objects for this cube/maps corresponding to the
                input coordinates. The length of the list is equal to the
                number of input coordinates.

        .. |spaxel| replace:: :class:`~marvin.tools.spaxel.Spaxel`

        """

        kwargs['cube'] = self.cube if spectrum else False
        kwargs['maps'] = self.get_unbinned()
        kwargs['modelcube'] = modelcube if not _is_MPL4(self._dapver) else False

        return marvin.utils.general.general.getSpaxel(x=x, y=y, ra=ra, dec=dec, **kwargs)

    def getMap(self, property_name, channel=None):
        """Retrieves a :class:`~marvin.tools.map.Map` object.

        Parameters:
            property_name (str):
                The property of the map to be extractred.
                E.g., `'emline_gflux'`.
            channel (str or None):
                If the ``property`` contains multiple channels,
                the channel to use, e.g., ``ha_6564'. Otherwise, ``None``.

        """

        return marvin.tools.map.Map(self, property_name, channel=channel)

    def getMapRatio(self, property_name, channel_1, channel_2):
        """Returns a ratio :class:`~marvin.tools.map.Map`.

        For a given ``property_name``, returns a :class:`~marvin.tools.map.Map`
        which is the ratio of ``channel_1/channel_2``.

        Parameters:
            property_name (str):
                The property_name of the map to be extractred.
                E.g., `'emline_gflux'`.
            channel_1,channel_2 (str):
                The channels to use.

        """

        # TODO extend to allow for different property names and make channel optional
        map_1 = self.getMap(property_name, channel=channel_1)
        map_2 = self.getMap(property_name, channel=channel_2)

        map_1.value /= map_2.value

        # TODO: do the error propogation (BHA)
        map_1.ivar = None

        map_1.mask &= map_2.mask

        map_1.channel = '{0}/{1}'.format(channel_1, channel_2)

        if map_1.unit != map_2.unit:
            map_1.unit = '{0}/{1}'.format(map_1.unit, map_2.unit)
        else:
            map_1.unit = ''

        return map_1

    def is_binned(self):
        """Returns True if the Maps is not unbinned."""

        if _is_MPL4(self._dapver):
            return self.bintype != __BINTYPES_MPL4_UNBINNED__
        else:
            return self.bintype != __BINTYPES_UNBINNED__

    def get_unbinned(self):
        """Returns a version of ``self`` corresponding to the unbinned Maps."""

        if _is_MPL4(self._dapver):
            unbinned_name = __BINTYPES_MPL4_UNBINNED__
        else:
            unbinned_name = __BINTYPES_UNBINNED__

        if self.bintype == unbinned_name:
            return self
        else:
            return Maps(plateifu=self.plateifu, release=self._release, bintype=unbinned_name,
                        template_kin=self.template_kin, template_pop=self.template_pop,
                        mode=self.mode)

    def get_bin_spaxels(self, binid, load=False, only_list=False):
        """Returns the list of spaxels belonging to a given ``binid``.

        If ``load=True``, the spaxel objects are loaded. Otherwise, they can be
        initiated by doing ``Spaxel.load()``. If ``only_list=True``, the method
        will return just a tuple containing the x and y coordinates of the spaxels.

        """

        if self.data_origin == 'file':
            spaxel_coords = zip(*np.where(self.data['BINID'].data.T == binid))

        elif self.data_origin == 'db':
            mdb = marvin.marvindb

            if _is_MPL4(self._dapver):
                table = mdb.dapdb.SpaxelProp
            else:
                table = mdb.dapdb.SpaxelProp5

            spaxel_coords = mdb.session.query(table.x, table.y).join(mdb.dapdb.File).filter(
                table.binid == binid, mdb.dapdb.File.pk == self.data.pk).order_by(
                    table.x, table.y).all()

        elif self.data_origin == 'api':
            url = marvin.config.urlmap['api']['getbinspaxels']['url']

            url_full = url.format(name=self.plateifu,
                                  bintype=self.bintype,
                                  template_kin=self.template_kin,
                                  binid=binid)

            try:
                response = self._toolInteraction(url_full)
            except Exception as ee:
                raise marvin.core.exceptions.MarvinError(
                    'found a problem requesting the spaxels for binid={0}: {1}'
                    .format(binid, str(ee)))

            response = response.getData()
            spaxel_coords = response['spaxels']

        spaxel_coords = list(spaxel_coords)
        if len(spaxel_coords) == 0:
            return []
        else:
            if only_list:
                return tuple([tuple(cc) for cc in spaxel_coords])

        spaxels = [marvin.tools.spaxel.Spaxel(x=cc[0], y=cc[1], maps=self, load=load)
                   for cc in spaxel_coords]

        return spaxels

    def get_bpt(self, method='kewley06', snr_min=3, return_figure=True,
                show_plot=True, use_oi=True, **kwargs):
        """Returns the BPT diagram for this target.

        This method produces the BPT diagram for this target using emission line maps and
        returns a dictionary of classification masks, that can be used to select spaxels
        that have been classified as belonging to a certain excitation process. It also
        provides plotting functionalities.

        Parameters:
            method ({'kewley06'}):
                The method used to determine the boundaries between different excitation
                mechanisms. Currently, the only available method is ``'kewley06'``, based on
                Kewley et al. (2006). Other methods may be added in the future. For a detailed
                explanation of the implementation of the method check the
                :ref:`BPT documentation <marvin-bpt>`.
            snr_min (float or dict):
                The signal-to-noise cutoff value for the emission lines used to generate the BPT
                diagram. If ``snr_min`` is a single value, that signal-to-noise will be used for
                all the lines. Alternatively, a dictionary of signal-to-noise values, with the
                emission line channels as keys, can be used.
                E.g., ``snr_min={'ha': 5, 'nii': 3, 'oi': 1}``. If some values are not provided,
                they will default to ``SNR>=3``.
            return_figure (bool):
                If ``True``, it also returns the matplotlib figure_ of the BPT diagram plot,
                which can be used to modify the style of the plot.
            show_plot (bool):
                If ``True``, interactively display the BPT plot.
            use_oi (bool):
                If ``True``, turns uses the OI diagnostic line in classifying BPT spaxels

        Returns:
            bpt_return:
                ``get_bpt`` always returns a dictionary of classification masks. These
                classification masks (not to be confused with bitmasks) are boolean arrays with the
                same shape as the Maps or Cube (without the spectral dimension) that can be used
                to select spaxels belonging to a certain excitation process (e.g., star forming).
                The keys of the dictionary, i.e., the classification categories, may change
                depending on the selected `method`. Consult the :ref:`BPT <marvin-bpt>`
                documentation for more details.
                If ``return_figure=True``, ``get_bpt`` will return a tuple, the first elemnt of
                which is the dictionary of classification masks, and the second the matplotlib
                figure.

        Example:
            >>> cube = Cube(plateifu='8485-1901')
            >>> maps = cube.getMaps()
            >>> bpt_masks, bpt_figure = maps.get_bpt(snr=5, return_figure=True, show_plot=False)

            Now we can use the masks to select star forming spaxels from the cube

            >>> sf_spaxels = cube.flux[bpt_masks['sf']['global']]

            And we can save the figure as a PDF

            >>> bpt_figure.savefig('8485_1901_bpt.pdf')

        .. _figure: http://matplotlib.org/api/figure_api.html

        """

        if 'snr' in kwargs:
            warnings.warn('snr is deprecated. Use snr_min instead. '
                          'snr will be removed in a future version of marvin',
                          marvin.core.exceptions.MarvinDeprecationWarning)
            snr_min = kwargs.pop('snr')
        elif len(kwargs.keys()) > 0:
            raise marvin.core.exceptions.MarvinError(
                'unknown keyword {0}'.format(list(kwargs.keys())[0]))

        # Makes sure all the keys in the snr keyword are lowercase
        if isinstance(snr_min, dict):
            snr_min = dict((kk.lower(), vv) for kk, vv in snr_min.items())

        # If we don't want the figure but want to show the plot, we still need to
        # temporarily get it.
        do_return_figure = True if return_figure or show_plot else False

        # Disables ion() if we are not showing the plot.
        plt_was_interactive = plt.isinteractive()
        if not show_plot and plt_was_interactive:
            plt.ioff()

        bpt_return = marvin.utils.dap.bpt.bpt_kewley06(self, snr_min=snr_min,
                                                       return_figure=do_return_figure,
                                                       use_oi=use_oi)

        if show_plot:
            plt.ioff()
            plt.show()

        # Restores original ion() status
        if plt_was_interactive and not plt.isinteractive():
            plt.ion()

        # Returs what we actually asked for.
        if return_figure and do_return_figure:
            return bpt_return
        elif not return_figure and do_return_figure:
            return bpt_return[0]
        else:
            return bpt_return

    def to_dataframe(self, columns=None, mask=None):
        ''' Converts the maps object into a Pandas dataframe

        Parameters:
            columns (list):
                The properties+channels you want to include.  Defaults to all of them.
            mask (array):
                A 2d mask array for filtering your data output

        Returns:
            df (DataFrame):
                a Pandas Dataframe
        '''

        allprops = list(itertools.chain(*[[p.fullname(c) for c in p.channels] if p.channels else [p.name] for p in self.properties]))
        if columns:
            allprops = [p for p in allprops if p in columns]
        data = np.array([self[p].value[mask].flatten() for p in allprops])

        # add a column for spaxel index
        spaxarr = np.array([np.where(mask.flatten())[0]]) if mask is not None else np.array([np.arange(data.shape[1])])
        data = np.concatenate((spaxarr, data), axis=0)
        allprops = ['spaxelid'] + allprops

        # create the dataframe
        df = pd.DataFrame(data.transpose(), columns=allprops)
        return df




from __future__ import print_function
import os
import requests
from brain.core.exceptions import BrainError, BrainApiAuthError, BrainNotImplemented
from brain import bconfig
from brain.core.core import URLMapDict
try:
    from urlparse import urljoin
except ImportError:
    from urllib.parse import urljoin
try:
    from cachecontrol import CacheControl as cache
except ImportError:
    cache = None

configkeys = []


class BrainInteraction(object):
    """ This class defines convenience wrappers for the Brain RESTful API """

    def __init__(self, route, params=None, request_type='post', auth='netrc',
                 timeout=(3.05, 300), headers=None):
        self.results = None
        self.route = route
        self.params = params
        self.request_type = request_type
        self.timeout = timeout
        self.headers = headers if headers is not None else {}
        self.statuscodes = {200: 'Ok', 401: 'Authentication Required', 404: 'URL Not Found',
                            500: 'Internal Server Error', 405: 'Method Not Allowed',
                            400: 'Bad Request', 502: 'Bad Gateway', 504: 'Gateway Timeout',
                            422: 'Unprocessable Entity'}

        self.url = urljoin(bconfig.sasurl, route) if self.route else None

        # set request Session
        self._setRequestSession()

        # set authentication
        self.setAuth(authtype=auth)

        # sends the request
        if self.url:
            self._sendRequest(request_type)
        else:
            raise BrainError('No route and/or url specified {0}'.format(self.url))

    def __repr__(self):
        return ('Interaction(route={0}, params={1}, request_type={2})'
                .format(self.route, repr(self.params), self.request_type))

    def _setRequestSession(self):
        ''' creates or sets the Brain requests Session object '''
        if isinstance(bconfig.request_session, type(None)):
            if not cache:
                bconfig.request_session = requests.Session()
            else:
                bconfig.request_session = cache(requests.Session())
            self.session = bconfig.request_session
        else:
            self.session = bconfig.request_session

    def _closeRequestSession(self):
        ''' closes the Brain requests Session object'''
        self.session.close()
        bconfig.request_session.close()
        bconfig.request_session = None

    def setAuth(self, authtype='netrc'):
        ''' set the session authentication '''
        self.authtype = authtype
        if authtype == 'netrc':
            # do nothing since this is default with no auth set
            pass
        elif authtype == 'http':
            from requests_toolbelt import GuessAuth
            auth = GuessAuth('user', 'passwd')
            self.session.auth = auth
        elif authtype == 'oauth':
            raise BrainNotImplemented('OAuth authentication')

    def _get_json_data(self, response):
        ''' Try to extract any json data '''
        try:
            json = response.json()
        except ValueError as e:
            json = None
        return json

    def _checkResponse(self, response):
        ''' Checks the response for proper http status code '''

        # check for bad status
        try:
            isbad = response.raise_for_status()
        except requests.HTTPError as http:
            self.status_code = response.status_code
            errmsg = 'Error accessing {0}: {1}-{2}'.format(response.url, response.status_code,
                                                           self.statuscodes[response.status_code])
            self.results = {'http_status_code': response.status_code, 'message': errmsg}
            json = self._get_json_data(response)
            if self.status_code == 401:
                if self.authtype == 'netrc':
                    msg = 'Please create or check credentials in your local .netrc file'
                elif self.authtype == 'oauth':
                    msg = 'Please check your Oauth authentication'
                elif self.authtype == 'http':
                    msg = 'Please check your http/digest authentication'
                self._closeRequestSession()
                raise BrainApiAuthError('API Authentication Error: {0}'.format(msg))
            elif self.status_code == 422:
                self._closeRequestSession()
                raise BrainError('Requests Http Status Error: {0}\nValidation Errors:\n{1}'.format(http, json))
            else:
                self._closeRequestSession()
                raise BrainError('Requests Http Status Error: {0}\n{1}'.format(http, json['api_error']['traceback']))
        else:
            # Not bad
            assert isbad is None, 'Http status code should not be bad'
            assert response.ok is True, 'Ok status should be true'

            # if 'route' not in self.url and 'clean' not in self.url:
            #     self._closeRequestSession()
            #     raise BrainError('test error is now raised here')

            self.status_code = response.status_code
            self.results = self._get_json_data(response)
            if not self.results:
                self.results = response.text
                self._closeRequestSession()
                raise BrainError('Response not in JSON format. {0}'.format(self.results))

            # Raises an error if status is -1
            if 'status' in self.results and self.results['status'] == -1:
                errorMsg = 'no error message provided' \
                    if 'error' not in self.results else self.results['error']
                self._check_for_traceback()
                self._closeRequestSession()
                raise BrainError('Something went wrong on the server side: {0}'.format(errorMsg))

    def _check_for_traceback(self):
        ''' checks the response for a traceback in the results response '''
        bconfig.traceback = self.results.get('traceback', None)

    def _sendRequest(self, request_type):
        ''' sends the api requests to the server '''
        assert request_type in ['get', 'post'], 'Valid request types are "get" and "post".'

        # Loads the local config parameters
        self._loadConfigParams()
        # Send the request
        try:
            if request_type == 'get':
                self._response = self.session.get(self.url, params=self.params, timeout=self.timeout, headers=self.headers)
            elif request_type == 'post':
                self._response = self.session.post(self.url, data=self.params, timeout=self.timeout, headers=self.headers)
        except requests.Timeout as rt:
            self._closeRequestSession()
            raise BrainError('Requests Timeout Error: {0}'.format(rt))
        except requests.URLRequired as urlreq:
            self._closeRequestSession()
            raise BrainError('Requests Valid URL Required: {0}'.format(urlreq))
        except requests.ConnectionError as con:
            self._closeRequestSession()
            raise BrainError('Requests Connection Error: {0}'.format(con))
        except requests.RequestException as req:
            self._closeRequestSession()
            raise BrainError('Ambiguous Requests Error: {0}'.format(req))
        else:
            # Check the response if it's good
            self._checkResponse(self._response)

    def getData(self, astype=None):
        data = self.results['data'] if 'data' in self.results else None

        if astype and data:
            try:
                return astype(data)
            except Exception as e:
                raise Exception('Failed: {0}, {1}'.format(e, data))
        else:
            return data

    # This should probably get moved into the server side code
    def _preloadResults(self):
        for key in configkeys:
            self.results[key] = bconfig.__getattribute__(key)

    def checkMyConfig(self):
        return {k: self.results[k] if k in self.results else '' for k in configkeys}

    def _loadConfigParams(self):
        """Load the local configuration into a parameters dictionary to be sent with the request"""
        self.params.update({'session_id': bconfig.session_id})

    def getRouteMap(self):
        """Retrieve the URL routing map if it exists."""

        if 'urlmap' in self.results:
            return URLMapDict(self.results['urlmap'])
        else:
            return URLMapDict()



#!/usr/bin/env python
# encoding: utf-8
#
# __init__.py
#
# Licensed under a 3-clause BSD license.
#

from __future__ import division
from __future__ import print_function
import os


class BrainConfig(object):

    def __init__(self):

        self.sasurl = os.getenv('SAS_URL') if 'SAS_URL' in os.environ else 'https://sas.sdss.org/'
        self._mode = 'auto'
        self.session_id = None
        self.request_session = None
        self.traceback = None

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, value):
        if value in ['local', 'remote', 'auto']:
            self._mode = value
        else:
            raise ValueError('config.mode must be "local" or "remote".')

bconfig = BrainConfig()



# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-02-12 17:38:51
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2017-06-23 13:55:06

from __future__ import print_function, division, absolute_import




#!/usr/bin/env python
# encoding: utf-8
#
# spaxel.py
#
# Licensed under a 3-clause BSD license.
#
# Revision history:
#     11 Apr 2016 J. Sánchez-Gallego
#       Initial version


from __future__ import division
from __future__ import print_function

import warnings

import numpy as np

import marvin
import marvin.core.core
import marvin.core.exceptions
import marvin.core.marvin_pickle
import marvin.utils.general.general
import marvin.tools.cube
import marvin.tools.maps
import marvin.tools.modelcube

from marvin.api import api
from marvin.core.exceptions import MarvinError, MarvinUserWarning, MarvinBreadCrumb
from marvin.tools.analysis_props import AnalysisProperty, DictOfProperties
from marvin.tools.spectrum import Spectrum

breadcrumb = MarvinBreadCrumb()


class Spaxel(object):
    """A class to interface with a spaxel in a cube.

    This class represents an spaxel with information from the reduced DRP
    spectrum, the DAP maps properties, and the model spectrum from the DAP
    logcube. A ``Spaxel`` can be initialised with all or only part of that
    information, and either from a file, a database, or remotely via the
    Marvin API. By definition a Marvin Spaxel is expected to be unbinned, i.e.,
    derived from a Maps and ModelCube that are unbinned themselves. For binned
    properties use ``Bin``. This behaviour can be overridden by using the
    ``allow_binned`` keyword.

    Parameters:
        x,y (int):
            The `x` and `y` coordinates of the spaxel in the cube (0-indexed).
        cube_filename (str):
            The path of the data cube file containing the spaxel to load.
        maps_filename (str):
            The path of the DAP Maps file containing the spaxel to load.
        modelcube_filename (str):
            The path of the DAP model cube file containing the spaxel to load.
        mangaid (str):
            The mangaid of the cube/maps of the spaxel to load.
        plateifu (str):
            The plate-ifu of the cube/maps of the spaxel to load (either
            ``mangaid`` or ``plateifu`` can be used, but not both).
        cube (:class:`~marvin.tools.cube.Cube` object or bool):
            If ``cube`` is a :class:`~marvin.tools.cube.Cube` object, that
            cube will be used for the ``Spaxel`` initilisitation. This mode
            is mostly intended for
            :class:`~marvin.utils.general.general.getSpaxel` as it
            significantly improves loading time. Otherwise, ``cube`` can be
            ``True`` (default), in which case a cube will be instantiated using
            the input ``filename``, ``mangaid``, or ``plateifu``. If
            ``cube=False``, no cube will be used and ``Spaxel.spectrum`` will
            not be populated.
        maps (:class:`~marvin.tools.maps.Maps` object or bool)
            As ``cube`` but populates ``Spaxel.properties`` with a dictionary
            of DAP measurements corresponding to the spaxel in the maps that
            matches ``bintype``, ``template_kin``, and ``template_pop``.
        modelcube (:class:`marvin.tools.modelcube.ModelCube` object or bool)
            As ``cube`` but populates ``Spaxel.model_flux``, ``Spaxel.model``,
            ``Spaxel.redcorr``, ``Spaxel.emline``, ``Spaxel.emline_base``, and
            ``Spaxel.stellar_continuum`` from the corresponding
            spaxel of the DAP modelcube that matches ``bintype``,
            ``template_kin``, and ``template_pop``.
        bintype (str or None):
            The binning type. For MPL-4, one of the following: ``'NONE',
            'RADIAL', 'STON'`` (if ``None`` defaults to ``'NONE'``).
            For MPL-5 and successive, one of, ``'ALL', 'NRE', 'SPX', 'VOR10'``
            (defaults to ``'ALL'``). Only allowed if ``allow_binned=True.```
        template_kin (str or None):
            The template use for kinematics. For MPL-4, one of
            ``'M11-STELIB-ZSOL', 'MILES-THIN', 'MIUSCAT-THIN'`` (if ``None``,
            defaults to ``'MIUSCAT-THIN'``). For MPL-5 and successive, the only
            option in ``'GAU-MILESHC'`` (``None`` defaults to it).
        template_pop (str or None):
            A placeholder for a future version in which stellar populations
            are fitted using a different template that ``template_kin``. It
            has no effect for now.
        release (str):
            The MPL/DR version of the data to use.
        load (bool):
            If ``True``, the spaxel data is loaded on initialisation. Otherwise,
            only the metadata is created. The spectra and properties can be then
            loaded by calling ``Spaxel.load()``.
        allow_binned (bool):
            If True, allows the spaxel to be instantiated from a binned combination
            of Maps and ModelCube.

    Attributes:
        spectrum (:class:`~marvin.tools.spectrum.Spectrum` object):
            A `Spectrum` object with the DRP spectrum and associated ivar and
            mask for this spaxel.
        specres (Numpy array):
            Median spectral resolution as a function of wavelength for the
            fibers in this IFU.
        specresd (Numpy array):
            Standard deviation of spectral resolution as a function of
            wavelength for the fibers in this IFU.
        properties (:class:`~marvin.tools.analysis_props.DictOfProperties`):
            A dotable, case-insensitive dictionary of
            :class:`~marvin.tools.analysis_props.AnalysisProperty` objects
            from the DAP maps extensions. The keys are a combination of
            category and channel, when applicable, e.g.,
            ``emline_sflux_siii_9533``.
        model_flux (:class:`~marvin.tools.spectrum.Spectrum` object):
            A `Spectrum` object with the flux of the binned spectrum. Includes
            ``ivar`` and ``mask``.
        wavelength (Numpy array):
            Wavelength vector, in Angstrom.
        redcorr (Numpy array):
            Reddening correction applied during the fitting procedures;
            ``dereddened_flux = model_flux.flux * redcorr``.
        model (:class:`~marvin.tools.spectrum.Spectrum` object):
            The best fitting model spectra (sum of the fitted continuum and
            emission-line models). Includes ``mask``.
        emline (:class:`~marvin.tools.spectrum.Spectrum` object):
            The model spectrum with only the emission lines. Includes ``mask``.
        emline_base (:class:`~marvin.tools.spectrum.Spectrum` object):
            The bitmask that only applies to the emission-line modeling.
            Includes ``mask``.

    """

    def __init__(self, *args, **kwargs):

        valid_kwargs = [
            'x', 'y', 'cube_filename', 'maps_filename', 'modelcube_filename',
            'mangaid', 'plateifu', 'cube', 'maps', 'modelcube', 'bintype',
            'template_kin', 'template_pop', 'release', 'load', 'allow_binned']

        assert len(args) == 0, 'Spaxel does not accept arguments, only keywords.'
        for kw in kwargs:
            assert kw in valid_kwargs, 'keyword {0} is not valid'.format(kw)

        self.__allow_binned = kwargs.pop('allow_binned', False)

        self.cube = kwargs.pop('cube', True) or False
        self.maps = kwargs.pop('maps', True) or False
        self.modelcube = kwargs.pop('modelcube', True) or False

        if not self.cube and not self.maps and not self.modelcube:
            raise MarvinError('either cube, maps, or modelcube must be True or '
                              'a Marvin Cube, Maps, or ModelCube object must be specified.')


        # drop breadcrumb
        breadcrumb.drop(message='Initializing MarvinSpaxel {0}'.format(self.__class__),
                        category=self.__class__)

        # Checks versions
        input_release = kwargs.pop('release', marvin.config.release)
        self._release = self._check_version(input_release)

        self._drpver, self._dapver = marvin.config.lookUpVersions(release=self._release)

        self.plateifu = None
        self.mangaid = None

        if len(args) > 0:
            self.x = int(args[0])
            self.y = int(args[1])
        else:
            self.x = int(kwargs.pop('x', None))
            self.y = int(kwargs.pop('y', None))

        assert self.x is not None and self.y is not None, 'Spaxel requires x and y to initialise.'

        self.loaded = False

        self.specres = None
        self.specresd = None
        self.spectrum = None
        self.properties = {}

        self.model_flux = None
        self.redcorr = None
        self.model = None
        self.emline = None
        self.emline_base = None
        self.stellar_continuum = None
        self._parent_shape = None

        self.plateifu = kwargs.pop('plateifu', None)
        self.mangaid = kwargs.pop('mangaid', None)

        self.bintype = None
        self.template_kin = None

        if self.maps or self.modelcube:

            # Some versions, like DR13, don't have an associated DAP, so we check.
            assert self._dapver, 'this MPL/DR version does not have an associated dapver.'

            self.bintype = marvin.tools.maps._get_bintype(
                self._dapver, bintype=kwargs.get('bintype', None))
            self.template_kin = marvin.tools.maps._get_template_kin(
                self._dapver, template_kin=kwargs.get('template_kin', None))

        self.__cube_filename = kwargs.pop('cube_filename', None)
        self.__maps_filename = kwargs.pop('maps_filename', None)
        self.__modelcube_filename = kwargs.pop('modelcube_filename', None)

        self._set_radec()

        if kwargs.pop('load', True):
            self.load()

    def _check_version(self, input_release):

        has_cube = isinstance(self.cube, marvin.tools.cube.Cube)
        has_maps = isinstance(self.maps, marvin.tools.maps.Maps)
        has_modelcube = isinstance(self.modelcube, marvin.tools.modelcube.ModelCube)

        if not has_cube and not has_maps and not has_modelcube:
            return input_release

        if has_cube and has_maps:

            assert self.cube._release == self.maps._release

            if has_modelcube:
                assert self.cube._release == self.modelcube._release

            return self.cube._release

        if has_cube and has_modelcube:
            assert self.cube._release == self.modelcube._release
            return self.cube._release

        if has_maps and has_modelcube:
            assert self.maps._release == self.modelcube._release
            return self.maps._release

        if has_cube:
            return self.cube._release

        if has_maps:
            return self.maps._release

        if has_modelcube:
            return self.modelcube._release

    def _set_radec(self):
        """Calculates ra and dec for this spaxel."""

        self.ra = None
        self.dec = None

        for obj in [self.cube, self.maps, self.modelcube]:
            if hasattr(obj, 'wcs'):
                if obj.wcs.naxis == 2:
                    self.ra, self.dec = obj.wcs.wcs_pix2world([[self.x, self.y]], 0)[0]
                elif obj.wcs.naxis == 3:
                    self.ra, self.dec, __ = obj.wcs.wcs_pix2world([[self.x, self.y, 0]], 0)[0]

    def load(self):
        """Loads the spaxel data."""

        if self.loaded:
            return

        self._check_cube()
        self._check_maps()
        self._check_modelcube()

        self.loaded = True

    def save(self, path, overwrite=False):
        """Pickles the spaxel to a file.

        This method will fail if any of ``cube``, ``maps``, or ``modelcube``
        has ``data_origin='db'``.

        Parameters:
            path (str):
                The path of the file to which the ``Spaxel`` will be saved.
                Unlike for other Marvin Tools that derive from
                :class:`~marvin.core.core.MarvinToolsClass`, ``path`` is
                mandatory for ``Spaxel`` given that the there is no default
                path for a given spaxel.
            overwrite (bool):
                If True, and the ``path`` already exists, overwrites it.
                Otherwise it will fail.

        Returns:
            path (str):
                The realpath to which the file has been saved.

        """

        return marvin.core.marvin_pickle.save(self, path=path, overwrite=overwrite)

    @classmethod
    def restore(cls, path, delete=False):
        """Restores a Spaxel object from a pickled file.

        If ``delete=True``, the pickled file will be removed after it has been
        unplickled. Note that, for objects with ``data_origin='file'``, the
        original file must exists and be in the same path as when the object
        was first created.

        """

        return marvin.core.marvin_pickle.restore(path, delete=delete)

    def _check_cube(self):
        """Loads the cube and the spectrum."""

        # Checks that the cube is correct or load ones if cube == True.
        if not isinstance(self.cube, bool):
            assert isinstance(self.cube, marvin.tools.cube.Cube), \
                'cube is not an instance of marvin.tools.cube.Cube or a boolean.'
        elif self.cube is True:
            self.cube = marvin.tools.cube.Cube(filename=self.__cube_filename,
                                               plateifu=self.plateifu,
                                               mangaid=self.mangaid,
                                               release=self._release)
        else:
            self.cube = None
            return

        if self.plateifu is not None:
            assert self.plateifu == self.cube.plateifu, \
                'input plateifu does not match the cube plateifu. '
        else:
            self.plateifu = self.cube.plateifu

        if self.mangaid is not None:
            assert self.mangaid == self.cube.mangaid, \
                'input mangaid does not match the cube mangaid. '
        else:
            self.mangaid = self.cube.mangaid

        self._parent_shape = self.cube.shape

        # Loads the spectrum
        self._load_spectrum()

    def _check_maps(self):
        """Loads the cube and the properties."""

        if not isinstance(self.maps, bool):
            assert isinstance(self.maps, marvin.tools.maps.Maps), \
                'maps is not an instance of marvin.tools.maps.Maps or a boolean.'
        elif self.maps is True:
            self.maps = marvin.tools.maps.Maps(filename=self.__maps_filename,
                                               mangaid=self.mangaid,
                                               plateifu=self.plateifu,
                                               template_kin=self.template_kin,
                                               release=self._release)
        else:
            self.maps = None
            return

        # Checks the bintype
        if self.maps.is_binned() and self.__allow_binned is False:
            raise MarvinError('cannot instantiate a Spaxel from a binned Maps.')

        if self.plateifu is not None:
            assert self.plateifu == self.maps.plateifu, \
                'input plateifu does not match the maps plateifu. '
        else:
            self.plateifu = self.maps.plateifu

        if self.mangaid is not None:
            assert self.mangaid == self.maps.mangaid, \
                'input mangaid does not match the maps mangaid. '
        else:
            self.mangaid = self.maps.mangaid

        self._parent_shape = self.maps.shape

        self.bintype = self.maps.bintype
        self.template_kin = self.maps.template_kin

        # Loads the properties
        self._load_properties()

    def _check_modelcube(self):
        """Loads the modelcube and associated arrays."""

        if not isinstance(self.modelcube, bool):
            assert isinstance(self.modelcube, marvin.tools.modelcube.ModelCube), \
                'modelcube is not an instance of marvin.tools.modelcube.ModelCube or a boolean.'
        elif self.modelcube is True:

            if self._is_MPL4():
                warnings.warn('ModelCube cannot be instantiated for MPL-4.',
                              MarvinUserWarning)
                self.modelcube = None
                return

            self.modelcube = marvin.tools.modelcube.ModelCube(filename=self.__modelcube_filename,
                                                              mangaid=self.mangaid,
                                                              plateifu=self.plateifu,
                                                              template_kin=self.template_kin,
                                                              release=self._release)
        else:
            self.modelcube = None
            return

        # Checks the bintype
        if self.modelcube.is_binned() and self.__allow_binned is False:
            raise MarvinError('cannot instantiate a Spaxel from a binned ModelCube.')

        self.bintype = self.modelcube.bintype
        self.template_kin = self.modelcube.template_kin

        if self.plateifu is not None:
            assert self.plateifu == self.modelcube.plateifu, \
                'input plateifu does not match the modelcube plateifu. '
        else:
            self.plateifu = self.modelcube.plateifu

        if self.mangaid is not None:
            assert self.mangaid == self.modelcube.mangaid, \
                'input mangaid does not match the modelcube mangaid. '
        else:
            self.mangaid = self.modelcube.mangaid

        self._parent_shape = self.modelcube.shape

        self._load_models()

    def __repr__(self):
        """Spaxel representation."""

        if not self.loaded:
            return '<Marvin Spaxel (x={0.x:d}, y={0.y:d}, loaded=False)'.format(self)

        # Gets the coordinates relative to the centre of the cube/maps.
        yMid, xMid = np.array(self._parent_shape) / 2.
        xCentre = int(self.x - xMid)
        yCentre = int(self.y - yMid)

        return ('<Marvin Spaxel (x={0.x:d}, y={0.y:d}; x_cen={1:d}, y_cen={2:d}>'.format(self,
                                                                                         xCentre,
                                                                                         yCentre))

    def _is_MPL4(self):
        """Returns True if the dapver correspond to MPL-4."""

        if self._dapver == '1.1.1':
            return True
        return False

    def _load_spectrum(self):
        """Initialises Spaxel.spectrum."""

        assert self.cube, 'a valid cube is needed to initialise the spectrum.'

        if self.cube.data_origin == 'file':

            cube_hdu = self.cube.data

            self.spectrum = Spectrum(cube_hdu['FLUX'].data[:, self.y, self.x],
                                     units='1E-17 erg/s/cm^2/Ang/spaxel',
                                     wavelength=cube_hdu['WAVE'].data,
                                     wavelength_unit='Angstrom',
                                     ivar=cube_hdu['IVAR'].data[:, self.y, self.x],
                                     mask=cube_hdu['MASK'].data[:, self.y, self.x])

            self.specres = cube_hdu['SPECRES'].data
            self.specresd = cube_hdu['SPECRESD'].data

        elif self.cube.data_origin == 'db':

            if marvin.marvindb is None:
                raise MarvinError('there is not a valid DB connection.')

            session = marvin.marvindb.session
            datadb = marvin.marvindb.datadb

            cube_db = self.cube.data

            spaxel = session.query(datadb.Spaxel).filter(
                datadb.Spaxel.cube == cube_db,
                datadb.Spaxel.x == self.x, datadb.Spaxel.y == self.y).one()

            if spaxel is None:
                raise MarvinError('cannot find an spaxel for x={0.x}, y={0.y}'.format(self))

            self.spectrum = Spectrum(spaxel.flux,
                                     units='1E-17 erg/s/cm^2/Ang/spaxel',
                                     wavelength=cube_db.wavelength.wavelength,
                                     wavelength_unit='Angstrom',
                                     ivar=spaxel.ivar,
                                     mask=spaxel.mask)

            self.specres = np.array(cube_db.specres)
            self.specresd = None

        elif self.cube.data_origin == 'api':

            # Calls the API to retrieve the DRP spectrum information for this spaxel.

            routeparams = {'name': self.plateifu, 'x': self.x, 'y': self.y}

            url = marvin.config.urlmap['api']['getSpectrum']['url'].format(**routeparams)

            # Make the API call
            response = api.Interaction(url, params={'release': self._release})

            # Temporarily stores the arrays prior to subclassing from np.array
            data = response.getData()

            # Instantiates the spectrum from the returned values from the Interaction
            self.spectrum = Spectrum(data['flux'],
                                     units='1E-17 erg/s/cm^2/Ang/spaxel',
                                     wavelength=data['wavelength'],
                                     wavelength_unit='Angstrom',
                                     ivar=data['ivar'],
                                     mask=data['mask'])

            self.specres = np.array(data['specres'])
            self.specresd = None

            return response

    def _load_properties(self):
        """Initialises Spaxel.properties."""

        assert self.maps, 'a valid maps is needed to initialise the properties.'

        maps_properties = self.maps.properties

        if self.maps.data_origin == 'file':

            maps_hdu = self.maps.data

            properties = {}
            for prop in maps_properties:

                prop_hdu = maps_hdu[prop.name]
                prop_hdu_ivar = None if not prop.ivar else maps_hdu[prop.name + '_ivar']
                prop_hdu_mask = None if not prop.mask else maps_hdu[prop.name + '_mask']

                if prop.channels:
                    for ii, channel in enumerate(prop.channels):

                        if isinstance(prop.unit, str) or not prop.unit:
                            unit = prop.unit
                        else:
                            unit = prop.unit[ii]

                        properties[prop.fullname(channel=channel)] = AnalysisProperty(
                            prop.name,
                            channel=channel,
                            value=prop_hdu.data[ii, self.y, self.x],
                            ivar=prop_hdu_ivar.data[ii, self.y, self.x] if prop_hdu_ivar else None,
                            mask=prop_hdu_mask.data[ii, self.y, self.x] if prop_hdu_mask else None,
                            unit=unit,
                            description=prop.description)

                else:

                    properties[prop.fullname(channel=channel)] = AnalysisProperty(
                        prop.name,
                        channel=None,
                        value=prop_hdu.data[self.y, self.x],
                        ivar=prop_hdu_ivar.data[self.y, self.x] if prop_hdu_ivar else None,
                        mask=prop_hdu_mask.data[self.y, self.x] if prop_hdu_mask else None,
                        unit=prop.unit,
                        description=prop.description)

        elif self.maps.data_origin == 'db':

            if marvin.marvindb is None:
                raise MarvinError('there is not a valid DB connection.')

            session = marvin.marvindb.session
            dapdb = marvin.marvindb.dapdb

            # Gets the spaxel_index for this spaxel.
            spaxel_index = self.x * self.maps.shape[0] + self.y

            spaxelprops_table = dapdb.SpaxelProp if self._is_MPL4() else dapdb.SpaxelProp5
            spaxelprops = session.query(spaxelprops_table).filter(
                spaxelprops_table.file == self.maps.data,
                spaxelprops_table.spaxel_index == spaxel_index).one()

            if spaxelprops is None:
                raise MarvinError('cannot find a spaxelprops for x={0.x}, y={0.y}'.format(self))

            properties = {}
            for prop in maps_properties:

                if prop.channels:

                    for ii, channel in enumerate(prop.channels):

                        if isinstance(prop.unit, str) or not prop.unit:
                            unit = prop.unit
                        else:
                            unit = prop.unit[ii]

                        properties[prop.fullname(channel=channel)] = AnalysisProperty(
                            prop.name,
                            channel=channel,
                            value=getattr(spaxelprops, prop.fullname(channel=channel)),
                            ivar=(getattr(spaxelprops, prop.fullname(channel=channel, ext='ivar'))
                                  if prop.ivar else None),
                            mask=(getattr(spaxelprops, prop.fullname(channel=channel, ext='mask'))
                                  if prop.mask else None),
                            unit=unit,
                            description=prop.description)

                else:

                    properties[prop.fullname()] = AnalysisProperty(
                        prop.name,
                        channel=None,
                        value=getattr(spaxelprops, prop.fullname()),
                        ivar=(getattr(spaxelprops, prop.fullname(ext='ivar'))
                              if prop.ivar else None),
                        mask=(getattr(spaxelprops, prop.fullname(ext='mask'))
                              if prop.mask else None),
                        unit=prop.unit,
                        description=prop.description)

        elif self.maps.data_origin == 'api':

            # Calls /api/<name>/properties/<path:path> to retrieve a
            # dictionary with all the properties for this spaxel.
            routeparams = {'name': self.plateifu,
                           'x': self.x, 'y': self.y,
                           'bintype': self.bintype,
                           'template_kin': self.template_kin}

            url = marvin.config.urlmap['api']['getProperties']['url'].format(**routeparams)

            # Make the API call
            response = api.Interaction(url, params={'release': self._release})

            # Temporarily stores the arrays prior to subclassing from np.array
            data = response.getData()

            properties = {}
            for prop_fullname in data['properties']:
                prop = data['properties'][prop_fullname]
                properties[prop_fullname] = AnalysisProperty(
                    prop['name'],
                    channel=prop['channel'],
                    value=prop['value'],
                    ivar=prop['ivar'],
                    mask=prop['mask'],
                    unit=prop['unit'],
                    description=prop['description'])

        self.properties = DictOfProperties(properties)

    def _load_models(self):

        assert self.modelcube, 'a ModelCube is needed to initialise models.'

        if self.modelcube.data_origin == 'file':

            hdus = self.modelcube.data
            flux_array = hdus['FLUX'].data[:, self.y, self.x]
            flux_ivar = hdus['IVAR'].data[:, self.y, self.x]
            mask = hdus['MASK'].data[:, self.y, self.x]
            model_array = hdus['MODEL'].data[:, self.y, self.x]
            model_emline = hdus['EMLINE'].data[:, self.y, self.x]
            model_emline_base = hdus['EMLINE_BASE'].data[:, self.y, self.x]
            model_emline_mask = hdus['EMLINE_MASK'].data[:, self.y, self.x]

        elif self.modelcube.data_origin == 'db':

            if marvin.marvindb is None:
                raise MarvinError('there is not a valid DB connection.')

            session = marvin.marvindb.session
            dapdb = marvin.marvindb.dapdb

            modelcube_db_spaxel = session.query(dapdb.ModelSpaxel).filter(
                dapdb.ModelSpaxel.modelcube == self.modelcube.data,
                dapdb.ModelSpaxel.x == self.x, dapdb.ModelSpaxel.y == self.y).one()

            if modelcube_db_spaxel is None:
                raise MarvinError('cannot find a modelcube spaxel for '
                                  'x={0.x}, y={0.y}'.format(self))

            flux_array = modelcube_db_spaxel.flux
            flux_ivar = modelcube_db_spaxel.ivar
            mask = modelcube_db_spaxel.mask
            model_array = modelcube_db_spaxel.model
            model_emline = modelcube_db_spaxel.emline
            model_emline_base = modelcube_db_spaxel.emline_base
            model_emline_mask = modelcube_db_spaxel.emline_mask

        elif self.modelcube.data_origin == 'api':

            # Calls /modelcubes/<name>/models/<path:path> to retrieve a
            # dictionary with all the models for this spaxel.
            url = marvin.config.urlmap['api']['getModels']['url']
            url_full = url.format(name=self.plateifu,
                                  bintype=self.bintype,
                                  template_kin=self.template_kin,
                                  x=self.x, y=self.y)

            try:
                response = api.Interaction(url_full, params={'release': self._release})
            except Exception as ee:
                raise MarvinError('found a problem when checking if remote model cube '
                                  'exists: {0}'.format(str(ee)))

            data = response.getData()

            flux_array = np.array(data['flux_array'])
            flux_ivar = np.array(data['flux_ivar'])
            mask = np.array(data['flux_mask'])
            model_array = np.array(data['model_array'])
            model_emline = np.array(data['model_emline'])
            model_emline_base = np.array(data['model_emline_base'])
            model_emline_mask = np.array(data['model_emline_mask'])

        # Instantiates the model attributes.

        self.redcorr = Spectrum(self.modelcube.redcorr,
                                wavelength=self.modelcube.wavelength,
                                wavelength_unit='Angstrom')

        self.model_flux = Spectrum(flux_array,
                                   units='1E-17 erg/s/cm^2/Ang/spaxel',
                                   wavelength=self.modelcube.wavelength,
                                   wavelength_unit='Angstrom',
                                   ivar=flux_ivar,
                                   mask=mask)

        self.model = Spectrum(model_array,
                              units='1E-17 erg/s/cm^2/Ang/spaxel',
                              wavelength=self.modelcube.wavelength,
                              wavelength_unit='Angstrom',
                              mask=mask)

        self.emline = Spectrum(model_emline,
                               units='1E-17 erg/s/cm^2/Ang/spaxel',
                               wavelength=self.modelcube.wavelength,
                               wavelength_unit='Angstrom',
                               mask=model_emline_mask)

        self.emline_base = Spectrum(model_emline_base,
                                    units='1E-17 erg/s/cm^2/Ang/spaxel',
                                    wavelength=self.modelcube.wavelength,
                                    wavelength_unit='Angstrom',
                                    mask=model_emline_mask)

        self.stellar_continuum = Spectrum(
            self.model.flux - self.emline.flux - self.emline_base.flux,
            units='1E-17 erg/s/cm^2/Ang/spaxel',
            wavelength=self.modelcube.wavelength,
            wavelength_unit='Angstrom',
            mask=model_emline_mask)

    @property
    def release(self):
        """Returns the release."""

        return self._release

    @release.setter
    def release(self, value):
        """Fails when trying to set the release after instatiation."""

        raise MarvinError('the release cannot be changed once the object has been instantiated.')



# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-04-08 14:59:40
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2017-04-09 09:07:34

from __future__ import print_function, division, absolute_import
from page_objects import PageObject, PageElement


class IndexPage(PageObject):
    imagepage = PageElement(id_='image_link')
    searchpage = PageElement(id_='search_link')


class SearchPage(PageObject):
    searchbox = PageElement(id_='searchbox')
    results = PageElement(id_='search_results')
    table = PageElement(id_='searchtablediv')




#!/usr/bin/env python3
# encoding: utf-8
#
# pickle.py
#
# Created by José Sánchez-Gallego on 7 Oct 2016.


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

try:
    import cPickle as pickle
except:
    import pickle

import warnings

import os

from marvin.utils.six import string_types
from marvin.core.exceptions import MarvinError, MarvinUserWarning


def save(obj, path=None, overwrite=False):
    """Pickles the object.

    If ``path=None``, uses the default location of the file in the tree
    but changes the extension of the file to ``.mpf`` (MaNGA Pickle File).
    Returns the path of the saved pickle file.
    
    Parameters:
        obj:
            Marvin object to pickle.
        path (str):
            Path of saved file. Default is ``None``.
        overwrite (bool):
            If ``True``, overwrite existing file. Default is ``False``.
    
    Returns:
        str:
            Path of saved file.
    """

    from marvin.core.core import MarvinToolsClass

    if path is None:
        assert isinstance(obj, MarvinToolsClass), 'path=None is only allowed for core objects.'
        path = obj._getFullPath()
        assert isinstance(path, string_types), 'path must be a string.'
        if path is None:
            raise MarvinError('cannot determine the default path in the '
                              'tree for this file. You can overcome this '
                              'by calling save with a path keyword with '
                              'the path to which the file should be saved.')

        # Replaces the extension (normally fits) with mpf
        if '.fits.gz' in path:
            path = path.strip('.fits.gz')
        else:
            path = os.path.splitext(path)[0]

        path += '.mpf'

    path = os.path.realpath(os.path.expanduser(path))

    if os.path.isdir(path):
        raise MarvinError('path must be a full route, including the filename.')

    if os.path.exists(path) and not overwrite:
        warnings.warn('file already exists. Not overwriting.', MarvinUserWarning)
        return

    dirname = os.path.dirname(path)
    if not os.path.exists(dirname):
        os.makedirs(dirname)

    try:
        with open(path, 'wb') as fout:
            pickle.dump(obj, fout, protocol=-1)
    except Exception as ee:
        if os.path.exists(path):
            os.remove(path)
        raise MarvinError('error found while pickling: {0}'.format(str(ee)))

    return path


def restore(path, delete=False):
    """Restores a MarvinToolsClass object from a pickled file.

    If ``delete=True``, the pickled file will be removed after it has been
    unplickled. Note that, for objects with ``data_origin='file'``, the
    original file must exists and be in the same path as when the object
    was first created.

    """

    assert isinstance(path, string_types), 'path must be a string.'

    path = os.path.realpath(os.path.expanduser(path))

    if os.path.isdir(path):
        raise MarvinError('path must be a full route, including the filename.')

    if not os.path.exists(path):
        raise MarvinError('the path does not exists.')

    try:
        with open(path, 'rb') as fin:
            obj = pickle.load(fin)
    except Exception as ee:
        raise MarvinError('something went wrong unplicking the object: {0}'
                          .format(str(ee)))

    if delete is True:
        os.remove(path)

    return obj



#!/usr/bin/env python
# encoding: utf-8

"""
Licensed under a 3-clause BSD license.

Revision History:
    Initial Version: 2016-02-17 17:46:57
    Last Modified On: 2016-02-17 17:46:57 by Brian

"""
from __future__ import print_function
from __future__ import division
from flask_classy import FlaskView
from flask import request
from brain import bconfig
from brain.core.exceptions import BrainError


def processRequest(request=None, as_dict=None, param=None):
    '''Generally process the request for POST or GET, and build a form dict

        Parameters:
            request (request):
                HTTP request object containing POST or GET data
            as_dict (bool):
                Boolean indicating whether to return the data as a standard dict or not
        Returns:
            Dict or ImmutableMultiDict
    '''

    # get form data
    if request.method == 'POST':
        if not request.form:
            # if data is content-type json
            data = request.get_json()
        else:
            # if data is content-type form
            data = request.form
    elif request.method == 'GET':
        data = request.args
    else:
        return {}

    # # if no data at all, return nothing
    if param and data:
        return data.get(param, None)

    # convert ImmutableMultiDict to dictionary (if get or post-form) or use dict if post-json
    if as_dict:
        if type(data) == dict:
            form = data
        else:
            # use multidict lists and iterlists to group multiple values for same in key into list
            try:
                # py2.7
                form = {key: val if len(val) > 1 else val[0] for key, val in data.iterlists()}
            except AttributeError:
                # py3.5
                form = {key: val if len(val) > 1 else val[0] for key, val in data.lists()}
    else:
        form = data

    return form


class BrainBaseView(FlaskView):
    """Super Class for all API Views to handle all global API items of interest"""

    def __init__(self):
        self.reset_results()
        bconfig.mode = 'local'

    def reset_results(self):
        self.results = {'data': None, 'status': -1, 'error': None, 'traceback': None}

    def update_results(self, newresults):
        self.results.update(newresults)

    def reset_status(self):
        self.results['status'] = -1

    def add_config(self):
        pass

    def before_request(self, *args, **kwargs):
        form = processRequest(request=request)
        self._release = form.get('release', None) if form else None
        self._endpoint = request.endpoint
        self.results['inconfig'] = form
        if form:
            for key, val in form.items():
                bconfig.__setattr__(key, val)
        # adds the out going config info into the results (placed here since didn't work in
        # after_request; obstensibly the in and out configs should match)
        self.add_config()

        # check API Authentication
        # self._checkAuth()

    def after_request(self, name, response):
        """This performs a reset of the results dict after every request method runs.

        See Flask-Classy for more info on after_request."""

        self.reset_results()
        return response

    def _checkAuth(self):
        ''' Checks the API for authentication '''

        print('api inconfig', self.results['inconfig'])
        if 'session_id' in self.results['inconfig']:
            session_id = self.results['inconfig'].get('session_id', None)
            # check for valid session_id
            if session_id is not None:
                # check if expired
                pass
            elif session_id is None or isexpired:
                # get new
                pass
        else:
            pass
            #raise BrainError('API session id not found in incoming request parameters')




# coding=utf-8
from __future__ import absolute_import

import sqlalchemy as sa

from tests import FormRelationsTestCase
from wtforms_alchemy import utils


class TestUtils(FormRelationsTestCase):
    def create_models(self):
        class Band(self.base):
            __tablename__ = 'band'
            id = sa.Column(sa.Integer, primary_key=True)
            name = sa.Column(sa.Unicode(255), nullable=False)

        class Person(self.base):
            __tablename__ = 'person'
            id = sa.Column(sa.Integer, primary_key=True)
            name = sa.Column(sa.Unicode(255), nullable=False)

        class BandMember(self.base):
            __tablename__ = 'band_members'
            band_id = sa.Column(sa.Integer, sa.ForeignKey(Band.id),
                                primary_key=True)
            band = sa.orm.relationship(Band, backref='members')
            person_id = sa.Column(sa.Integer, sa.ForeignKey(Person.id),
                                  primary_key=True)
            person = sa.orm.relationship(Person,
                                         backref=sa.orm.backref('band_role',
                                                                uselist=False))
            role = sa.Column(sa.Unicode(255), nullable=False)

            def __repr__(self):
                fmt = '<BandMember band_id={} person_id={} at {:x}>'
                return fmt.format(repr(self.band_id), repr(self.person_id),
                                  id(self))

        self.Band = Band
        self.Person = Person
        self.BandMember = BandMember

    def create_forms(self):
        pass

    def test_find_entity(self):
        band = self.Band(name=u'The Furious')
        self.session.add(band)
        singer = self.Person(name=u'Paul Insane')
        self.BandMember(band=band, person=singer, role=u'singer')
        guitarist = self.Person(name=u'John Crazy')
        self.BandMember(band=band, person=guitarist, role=u'guitar')
        self.session.commit()

        sing_data = dict(band_id=band.id, person_id=singer.id, role=u'singer')
        guitar_data = dict(
            band_id=band.id,
            person_id=guitarist.id,
            role=u'guitar'
        )

        assert (utils.find_entity(band.members, self.BandMember, sing_data)
                is singer.band_role)
        assert (utils.find_entity(band.members, self.BandMember, guitar_data)
                is guitarist.band_role)



#!/usr/bin/env python
# encoding: utf-8
#
# Licensed under a 3-clause BSD license.
#
#
# map.py
#
# Created by José Sánchez-Gallego on 26 Jun 2016.


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from distutils import version
import os
import warnings

from astropy.io import fits
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm

import marvin
import marvin.api.api
import marvin.core.marvin_pickle
import marvin.core.exceptions
import marvin.tools.maps
import marvin.utils.plot.map


try:
    import sqlalchemy
except ImportError:
    sqlalchemy = None


class Map(object):
    """Describes a single DAP map in a Maps object.

    Unlike a ``Maps`` object, which contains all the information from a DAP
    maps file, this class represents only one of the multiple 2D maps contained
    within. For instance, ``Maps`` may contain emission line maps for multiple
    channels. A ``Map`` would be, for example, the map for ``emline_gflux`` and
    channel ``ha_6564``.

    A ``Map`` is basically a set of three Numpy 2D arrays (``value``, ``ivar``,
    and ``mask``), with extra information and additional methods for
    functionality.

    ``Map`` objects are not intended to be initialised directly, at least
    for now. To get a ``Map`` instance, use the
    :func:`~marvin.tools.maps.Maps.getMap` method.

    Parameters:
        maps (:class:`~marvin.tools.maps.Maps` object):
            The :class:`~marvin.tools.maps.Maps` instance from which we
            are extracting the ``Map``.
        property_name (str):
            The category of the map to be extractred. E.g., `'emline_gflux'`.
        channel (str or None):
            If the ``property`` contains multiple channels, the channel to use,
            e.g., ``ha_6564'. Otherwise, ``None``.

    """

    def __init__(self, maps, property_name, channel=None):

        assert isinstance(maps, marvin.tools.maps.Maps)

        self.maps = maps
        self.property_name = property_name.lower()
        self.channel = channel.lower() if channel else None
        self.shape = self.maps.shape

        self.release = maps.release

        self.maps_property = self.maps.properties[self.property_name]
        if (self.maps_property is None or
                (self.maps_property.channels is not None and
                 self.channel not in self.maps_property.channels)):
            raise marvin.core.exceptions.MarvinError(
                'invalid combination of property name and channel.')

        self.value = None
        self.ivar = None
        self.mask = None

        self.header = None
        self.unit = None

        if maps.data_origin == 'file':
            self._load_map_from_file()
        elif maps.data_origin == 'db':
            self._load_map_from_db()
        elif maps.data_origin == 'api':
            self._load_map_from_api()

        self.masked = np.ma.array(self.value, mask=(self.mask > 0
                                                    if self.mask is not None else False))

    def __repr__(self):

        return ('<Marvin Map (plateifu={0.maps.plateifu!r}, property={0.property_name!r}, '
                'channel={0.channel!r})>'.format(self))

    @property
    def snr(self):
        """Returns the signal-to-noise ratio for each spaxel in the map."""

        return np.abs(self.value * np.sqrt(self.ivar))

    def _load_map_from_file(self):
        """Initialises the Map from a ``Maps`` with ``data_origin='file'``."""

        self.header = self.maps.data[self.property_name].header

        if self.channel is not None:
            channel_idx = self.maps_property.channels.index(self.channel)
            self.value = self.maps.data[self.property_name].data[channel_idx]
            if self.maps_property.ivar:
                self.ivar = self.maps.data[self.property_name + '_ivar'].data[channel_idx]
            if self.maps_property.mask:
                self.mask = self.maps.data[self.property_name + '_mask'].data[channel_idx]
        else:
            self.value = self.maps.data[self.property_name].data
            if self.maps_property.ivar:
                self.ivar = self.maps.data[self.property_name + '_ivar'].data
            if self.maps_property.mask:
                self.mask = self.maps.data[self.property_name + '_mask'].data

        if isinstance(self.maps_property.unit, list):
            self.unit = self.maps_property.unit[channel_idx]
        else:
            self.unit = self.maps_property.unit

        return

    def _load_map_from_db(self):
        """Initialises the Map from a ``Maps`` with ``data_origin='db'``."""

        mdb = marvin.marvindb

        if not mdb.isdbconnected:
            raise marvin.core.exceptions.MarvinError('No db connected')

        if sqlalchemy is None:
            raise marvin.core.exceptions.MarvinError('sqlalchemy required to access the local DB.')

        if version.StrictVersion(self.maps._dapver) <= version.StrictVersion('1.1.1'):
            table = mdb.dapdb.SpaxelProp
        else:
            table = mdb.dapdb.SpaxelProp5

        fullname_value = self.maps_property.fullname(channel=self.channel)
        value = mdb.session.query(getattr(table, fullname_value)).filter(
            table.file_pk == self.maps.data.pk).order_by(table.spaxel_index).all()
        self.value = np.array(value).reshape(self.shape).T

        if self.maps_property.ivar:
            fullname_ivar = self.maps_property.fullname(channel=self.channel, ext='ivar')
            ivar = mdb.session.query(getattr(table, fullname_ivar)).filter(
                table.file_pk == self.maps.data.pk).order_by(table.spaxel_index).all()
            self.ivar = np.array(ivar).reshape(self.shape).T

        if self.maps_property.mask:
            fullname_mask = self.maps_property.fullname(channel=self.channel, ext='mask')
            mask = mdb.session.query(getattr(table, fullname_mask)).filter(
                table.file_pk == self.maps.data.pk).order_by(table.spaxel_index).all()
            self.mask = np.array(mask).reshape(self.shape).T

        # Gets the header
        hdus = self.maps.data.hdus
        header_dict = None
        for hdu in hdus:
            if self.maps_property.name.upper() == hdu.extname.name.upper():
                header_dict = hdu.header_to_dict()
                break

        if not header_dict:
            warnings.warn('cannot find the header for property {0}.'
                          .format(self.maps_property.name),
                          marvin.core.exceptions.MarvinUserWarning)
        else:
            self.header = fits.Header(header_dict)

        self.unit = self.maps_property.unit

    def _load_map_from_api(self):
        """Initialises the Map from a ``Maps`` with ``data_origin='api'``."""

        url = marvin.config.urlmap['api']['getmap']['url']

        url_full = url.format(
            **{'name': self.maps.plateifu,
               'property_name': self.property_name,
               'channel': self.channel,
               'bintype': self.maps.bintype,
               'template_kin': self.maps.template_kin})

        try:
            response = marvin.api.api.Interaction(url_full,
                                                  params={'release': self.maps._release})
        except Exception as ee:
            raise marvin.core.exceptions.MarvinError(
                'found a problem when getting the map: {0}'.format(str(ee)))

        data = response.getData()

        if data is None:
            raise marvin.core.exceptions.MarvinError(
                'something went wrong. Error is: {0}'.format(response.results['error']))

        self.value = np.array(data['value'])
        self.ivar = np.array(data['ivar']) if data['ivar'] is not None else None
        self.mask = np.array(data['mask']) if data['mask'] is not None else None
        self.unit = data['unit']
        self.header = fits.Header(data['header'])

        return

    def save(self, path, overwrite=False):
        """Pickles the map to a file.

        This method will fail if the map is associated to a Maps loaded
        from the db.

        Parameters:
            path (str):
                The path of the file to which the ``Map`` will be saved.
                Unlike for other Marvin Tools that derive from
                :class:`~marvin.core.core.MarvinToolsClass`, ``path`` is
                mandatory for ``Map`` given that the there is no default
                path for a given map.
            overwrite (bool):
                If True, and the ``path`` already exists, overwrites it.
                Otherwise it will fail.

        Returns:
            path (str):
                The realpath to which the file has been saved.

        """

        # check for file extension
        if not os.path.splitext(path)[1]:
            path = os.path.join(path + '.mpf')

        return marvin.core.marvin_pickle.save(self, path=path, overwrite=overwrite)

    @classmethod
    def restore(cls, path, delete=False):
        """Restores a Map object from a pickled file.

        If ``delete=True``, the pickled file will be removed after it has been
        unplickled. Note that, for map objects instantiated from a Maps object
        with ``data_origin='file'``, the original file must exists and be
        in the same path as when the object was first created.

        """

        return marvin.core.marvin_pickle.restore(path, delete=delete)

    def plot(self, *args, **kwargs):
        """Convenience method that wraps :func:`marvin.utils.plot.map.plot`.
        
        See :func:`marvin.utils.plot.map.plot` for full documentation.
        """
        return marvin.utils.plot.map.plot(dapmap=self, *args, **kwargs)



# Copyright 2015 SolidBuilds.com. All rights reserved.
#
# Authors: Ling Thio <ling.thio@gmail.com>

from .conftest import the_db as db


class GrandParent(db.Model):
    __tablename__ = 'grandparents'
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(50), nullable=False, server_default='')


class Parent(db.Model):
    __tablename__ = 'parents'
    id = db.Column(db.Integer, primary_key=True)
    grandparent_id = db.Column(db.Integer(), db.ForeignKey('grandparents.id', ondelete='CASCADE'))
    grandparent = db.relationship('GrandParent')
    name = db.Column(db.String(50), nullable=False, server_default='')


class Record(db.Model):
    __tablename__ = 'records'
    id = db.Column(db.Integer, primary_key=True)
    parent_id = db.Column(db.Integer(), db.ForeignKey('parents.id', ondelete='CASCADE'))
    parent = db.relationship('Parent')

    string = db.Column(db.String(50), nullable=False, server_default='')
    unicode = db.Column(db.Unicode(50), nullable=False, server_default=u'')
    boolean = db.Column(db.Boolean(), nullable=False, server_default='0')
    integer = db.Column(db.Integer(), nullable=False, server_default='0')
    float = db.Column(db.Float(), nullable=False, server_default='0.0')





from tests import ModelFormTestCase, MultiDict
from wtforms_alchemy import ModelForm


class TestStringFieldTrimming(ModelFormTestCase):
    def test_strip_string_fields_set_for_string_field(self):
        self.init()

        class ModelTestForm(ModelForm):
            class Meta:
                model = self.ModelTest
                strip_string_fields = True

        f = ModelTestForm(MultiDict([('test_column', 'strip this   ')]))
        assert f.test_column.data == 'strip this'

    def test_does_not_trim_fields_when_trim_param_is_false(self):
        self.init(info={'trim': False})

        class ModelTestForm(ModelForm):
            class Meta:
                model = self.ModelTest
                strip_string_fields = True

        f = ModelTestForm(MultiDict([('test_column', 'strip this   ')]))
        assert f.test_column.data == 'strip this   '
